# do not import all endpoints into this module because that uses a lot of memory and stack frames
# if you need the ability to import all endpoints from this module, import them with
# from sedaro_base_client.paths.simulations_jobs_job_id_externals_agent_id_external_state_block_id import Api

from sedaro_base_client.paths import PathValues

path = PathValues.SIMULATIONS_JOBS_JOB_ID_EXTERNALS_AGENT_ID_EXTERNAL_STATE_BLOCK_ID


import datetime as dt
from pathlib import Path
from typing import List

from sedaro.branches.scenario_branch.sim_client import Simulation
from sedaro.results.simulation_result import SimulationResult
from sedaro.results.utils import HFILL, STATUS_ICON_MAP, hfill
from sedaro.settings import STATUS


class StudyResult:

    def __init__(self, study, metadata: dict):
        '''Initialize a new Study Result.

        By default, this class will lazily load simulation results as requested
        and cache them in-memory. Different caching options can be enabled with
        the .set_cache method.
        '''
        self.__study = study
        self.__metadata = metadata
        self.__cache = True
        self.__cache_dir = None
        self.__cached_sim_results = {}

    def __repr__(self) -> str:
        return f'SedaroStudyResult(branch={self.branch}, status={self.status}, iterations={self.iterations})'

    def __iter__(self):
        '''Iterate through simulation results.'''
        return (self.result(id_) for id_ in self.job_ids)

    def __contains__(self, id_):
        '''Check if this study contains a certain simulation ID.'''
        return id_ in self.job_ids

    @property
    def id(self) -> int:
        return self.__metadata['id']

    @property
    def branch(self) -> str:
        return self.__metadata['branch']

    @property
    def scenario_hash(self) -> str:
        return self.__metadata['scenarioHash']

    @property
    def status(self) -> str:
        return self.__metadata[STATUS]

    @property
    def date_created(self) -> dt.datetime:
        return dt.datetime.fromisoformat(str(self.__metadata['dateCreated'])[:-1])

    @property
    def date_modified(self) -> dt.datetime:
        return dt.datetime.fromisoformat(str(self.__metadata['dateModified'])[:-1])

    @property
    def job_ids(self) -> List[int]:
        return [entry for entry in self.__metadata['jobs']]

    @property
    def iterations(self) -> int:
        return len(self.__metadata['jobs'])

    def set_cache(self, cache: bool = True, cache_dir: str = None) -> None:
        '''Set caching options for this study result.

        Args:
            cache: Boolean option to turn caching on or off.
            cache_dir: Path to a directory for on-disk caching.
        '''
        self.__cache = cache
        self.__cache_dir = cache_dir
        if self.__cache:
            if cache_dir is not None:
                self.__cache_dir = Path(cache_dir)
                if not self.__cache_dir.is_dir():
                    raise ValueError(f'Cache directory not found: {cache_dir}')
        else:
            if self.__cache_dir is not None:
                raise ValueError(f'Cache directory is defined but caching is off.')
        self.__cached_sim_results = {}

    def result(self, id_: str) -> SimulationResult:
        '''Query results for a particular simulation.'''
        result = None
        if self.__cache:
            if self.__cache_dir is None:
                if id_ in self.__cached_sim_results:
                    result = self.__cached_sim_results[id_]
            else:
                cache_file = self.__cache_dir / f'sim_{id_}.cache'
                if cache_file.exists():
                    result = SimulationResult.from_file(cache_file)

        if result is None:
            print(f'💾 Downloading simulation result id {id_}...', end='')
            result = Simulation(self.__study._sedaro, self.__study._branch).results(id_)
            print('done!')

        if self.__cache:
            if self.__cache_dir is None:
                self.__cached_sim_results[id_] = result
            else:
                if not cache_file.exists():
                    result.to_file(cache_file)

        return result

    def clear_cache(self) -> None:
        self.__cached_sim_results = {}
        if self.__cache_dir is not None:
            for file in self.__cache_dir.glob('sim_*.cache'):
                file.unlink()

    def refresh(self) -> None:
        '''Update metadata for this study.'''
        from sedaro.branches.scenario_branch.study_client import Study
        self.__metadata = self.__study.status(self.id)

    def summarize(self) -> None:
        '''Summarize these results in the console.'''
        hfill()
        print(f'Sedaro Study Result Summary'.center(HFILL))
        print(f'Job ID {self.id}'.center(HFILL))
        hfill()
        print(f'{STATUS_ICON_MAP[self.status]} Study {self.status.lower()}')

        # print(f'\n📇 Scenario identification hash: {self.scenario_hash[:12]}')

        if self.iterations == 1:
            print('\n📋 Study contains 1 simulation')
        else:
            print(f'\n📋 Study contains {self.iterations} simulations')
        print(f'\n {str(self.job_ids)}')

        if self.__cache:
            if self.__cache_dir is None:
                print(f'\n❗ In-memory simulation result caching is ON')
            else:
                print(f'\n💾 Caching data to: {self.__cache_dir}')

        hfill()
        print("❓ Query individual simulation results with .result(<ID>)")



import json
import math
from os import listdir
from pathlib import Path
from typing import Union

DEFAULT_HOST = 'https://api.sedaro.com'
ENGINE_MAP = {
    '0': 'gnc',
    '1': 'cdh',
    '2': 'power',
    '3': 'thermal',
}
ENGINE_MAP_REVERSED = {
    'gnc': '0',
    'cdh': '1',
    'power': '2',
    'thermal': '3',
}
ENGINE_MAP_CASED = {
    'gnc': 'GNC',
    'cdh': 'CDH',
    'power': 'Power',
    'thermal': 'Thermal',
}
ENGINE_EXPANSION = {
    'gnc': 'Guidance, Navigation, & Control',
    'cdh': 'Command & Data Handling',
    'power': 'Power',
    'thermal': 'Thermal',
}
STATUS_ICON_MAP = {
    "SUCCEEDED": "✅",
    "FAILED": "❌",
    "TERMINATED": "❌",
    "PAUSED": "⏸️",
    "PENDING": "⌛",
    "PROVISIONING": "⌛",
    "CONFIGURING": "⌛",
    "BUILDING": "⌛",
    "RUNNING": "⌛",
    "ERROR": "❌"
}
HFILL = 75


def hfill(char="-", len=HFILL):
    print(char * len)


def _element_id_dict(agent_data):
    '''Break out all blocks into a dict where each key is an ID.'''
    out = {}
    if 'blocks' in agent_data:
        for id_, value in agent_data['blocks'].items():
            if 'id' in value:
                if id_ in out:
                    raise ValueError(f"Duplicate ID {id_}")
                else:
                    out[id_] = value
    return out


def _block_type_in_supers(block_type: str, meta_supers: dict, super_type: str = 'Agent') -> bool:
    if block_type == super_type:
        return True
    elif block_type in meta_supers:
        supertypes = meta_supers[block_type]
        if len(supertypes) == 0:
            return False
        return any(_block_type_in_supers(supertype, meta_supers, super_type=super_type) for supertype in supertypes)
    else:
        return False


def _get_agent_id_name_map(meta):
    '''Get mapping from agent ID to name.'''
    return {
        id_: entry['name']
        for id_, entry in meta['structure']['scenario']['blocks'].items()
        if _block_type_in_supers(entry['type'], meta['structure']['scenario']['_supers'])
    }


def _restructure_data(series, agents, meta):
    agent_mapping = {}
    blocks = {}
    index = {}

    for series_key in series:
        agent_id = series_key.split("/")[0]
        try:
            agent_mapping[agent_id] = agents[agent_id]
        except KeyError:  # agent_id corresponding to coupling is not in the metamodel
            continue
        if agent_id not in blocks:
            blocks[agent_id] = _element_id_dict(meta['structure']['agents'].get(agent_id, {}))
        if agent_id not in index:
            index[agent_id] = {}
        df = series[series_key]

        columns = df.columns.tolist()
        for column in columns:
            if '/' not in column:  # ignore engine variables
                elements = column.split(".")
                first_element = elements[0]
                if first_element not in blocks[agent_id]:
                    block_id = 'root'
                    path_components = elements
                else:
                    block_id = first_element
                    path_components = elements[1:]
                if block_id not in index[agent_id]:
                    index[agent_id][block_id] = {}
                this_block_index = index[agent_id][block_id]
                if series_key not in this_block_index:
                    this_block_index[series_key] = {}
                ptr = this_block_index[series_key]
                for element in path_components:
                    if element not in ptr:
                        ptr[element] = {}
                    ptr = ptr[element]

    return agent_mapping, blocks, index


def _get_series_type(series):
    for entry in series:
        if entry is not None:
            return type(entry).__name__
    else:
        return "None"


def bsearch(ordered_series, value):
    '''Binary search for a value in an ordered series.

    Returns the index of the value in the series, or the index of the immediately
    lower value if the value is not present.
    '''
    def _bsearch(low, high):
        if high == low:
            return low
        mid = math.ceil((high + low) / 2)
        if ordered_series[mid] == value:
            return mid
        elif ordered_series[mid] > value:
            return _bsearch(low, mid - 1)
        else:
            return _bsearch(mid, high)
    if value < ordered_series[0]:
        return -1
    return _bsearch(0, len(ordered_series) - 1)


def get_column_names(column_index, prefix):
    """
    For example:
        column_index = {'body_eci': {'0': {}, '1': {}, '2': {}, '3': {}}, 'body_ecef': {'0': {}, '1': {}, '2': {}, '3': {}}}
        prefix = 'attitude'
    Returns:
        [
            'attitude.body_eci.0',
            'attitude.body_eci.1',
            'attitude.body_eci.2',
            'attitude.body_eci.3',
            'attitude.body_ecef.0',
            'attitude.body_ecef.1',
            'attitude.body_ecef.2',
            'attitude.body_ecef.3'
        ]
    """
    if len(column_index) == 0:
        return [prefix]
    else:
        # don't add a dot if prefix is empty (this is the case for the root block)
        if len(prefix) > 0:
            prefix = f"{prefix}."
        columns = []
        for key in column_index:
            columns.extend(get_column_names(column_index[key], f"{prefix}{key}"))
        return columns


def get_parquets(path: str):
    paths = listdir(path)
    return [parquet for parquet in paths if not parquet.startswith('.')]


VLLS = [
    'visibleEarthArea',
    'activeRoutines',
    'availableTransmitters',
    'pseudoranges',
]


def parse_set_string(set_string):
    if set_string is None:
        return None
    # Converts a set represented as a string to a list
    values = set_string.strip("{}").split(",")
    return [item.strip().strip("'") for item in values]


def values_from_df(values, name=None):
    if not name:
        return values
    if name == 'availableTransmitters' or ('.' in name and name.split('.')[-1] == 'availableTransmitters'):
        # data for availableTransmitters is returned as a list of sets (represented as strings) which is not JSON-serializable
        return [parse_set_string(s) for s in values]
    elif name in VLLS or ('.' in name and name.split('.')[-1] in VLLS):
        return [json.loads(item.replace("'", '"')) for item in values]
    else:
        return values


def value_from_df(value, name=None):
    if not name:
        return value
    if name == 'availableTransmitters' or ('.' in name and name.split('.')[-1] == 'availableTransmitters'):
        # data for availableTransmitters is returned as a list of sets (represented as strings) which is not JSON-serializable
        return parse_set_string(value)
    elif name in VLLS or ('.' in name and name.split('.')[-1] in VLLS):
        return json.loads(value.replace("'", '"'))
    else:
        return value


def get_static_data(static_data: dict, object_type: str, engine: str = None):
    if len(static_data) == 0:
        raise ValueError(f"No static data available for this {object_type}.")
    elif engine is None:
        return static_data
    else:
        prefix = list(static_data.keys())[0][:-1]
        if engine.lower() in ENGINE_MAP_REVERSED.keys():
            stream_int = ENGINE_MAP_REVERSED[engine.lower()]
            try:
                return static_data[f"{prefix}{stream_int}"]
            except KeyError:
                raise KeyError(f"No static data available for the specified engine for this {object_type}.")
        else:
            raise ValueError(f"{engine} is not a valid engine identifier.")


def get_static_data_engines(static_data: dict):
    if static_data is None:
        return []
    else:
        return [ENGINE_MAP_CASED[ENGINE_MAP[stream_id[-1]]] for stream_id in static_data.keys()]


class FromFileAndToFileAreDeprecated:
    @classmethod
    def from_file(self, filename: Union[str, Path]):
        print("Warning: `from_file` is deprecated. Use `load` instead. Calling `load`.")
        return self.load(filename)

    def to_file(self, filename: Union[str, Path]):
        print("Warning: `to_file` is deprecated. Use `save` instead. Calling `save`.")
        return self.save(filename)



import os


PACKAGE_NAME = 'sedaro'
BASE_PACKAGE_NAME = f'{PACKAGE_NAME}_base_client'

COMMON_API_KWARGS = {
    'skip_deserialization': True
}
"""To be spread in autogenerated HTTP request methods. Make sure to use in conjunction with `body_from_res`."""


# template
CRUD = 'crud'
'''One of the keys in returned response from a template crud PATCH request'''
BLOCKS = 'blocks'
INDEX = 'index'
RELATIONSHIPS = '_relationships'
SCENARIO = 'Scenario'
LIBRARY = 'Library'

ROOT = 'root'

# Relationships and Blocks
MANY_SIDE = 'ManySide'
DATA_SIDE = 'DataSide'
ONE_SIDE = 'OneSide'
TYPE = 'type'
ID = 'id'

# Statuses
STATUS = 'status'

PENDING = 'PENDING'
QUEUED = 'QUEUED'
PROVISIONING = 'PROVISIONING'
CONFIGURING = 'CONFIGURING'
BUILDING = 'BUILDING'

RUNNING = 'RUNNING'

TERMINATED = 'TERMINATED'
SUCCEEDED = 'SUCCEEDED'

FAILED = 'FAILED'
ERROR = 'ERROR'

PRE_RUN_STATUSES = {QUEUED, PENDING, PROVISIONING, CONFIGURING, BUILDING}
BAD_STATUSES = {FAILED, ERROR}

GRPC_MAX_MESSAGE_LENGTH = 50 * 1024 * 1024 # 50 MB



import json
from typing import TYPE_CHECKING, Literal, Optional, Union, overload

import requests
from urllib3.response import HTTPResponse

from .utils import check_for_res_error, parse_urllib_response

if TYPE_CHECKING:
    from .sedaro_api_client import SedaroApiClient


class PlainRequest:
    def __init__(self, sedaro: 'SedaroApiClient') -> None:
        self.__sedaro = sedaro

    def __request(
        self,
        resource_path: str,
        method: str,
        body: Optional[dict] = None,
        raw: bool = False,
    ):
        """Send a request to the Sedaro server

        Args:
            resource_path (str): url path (everything after the host) for desired route
            method (str): HTTP method ('GET', 'POST', 'DELETE'...etc)
            body (Optional[dict], optional): Body of the request. Defaults to None.
            raw (optional, bool): triggers returning the raw response rather than a dictionary of the response data.\
                Defaults to `False`.

        Returns:
            dict: dictionary from the response body
        """
        headers = {}

        if body is not None:
            body = json.dumps(body)
            headers['Content-Type'] = 'application/json'
            headers['Accept'] = 'application/json'  # Required for Sedaro firewall

        with self.__sedaro.api_client() as api:
            res = api.call_api(
                resource_path,
                method,
                headers=headers,
                body=body
            )
        if raw:
            return res

        res_ = parse_urllib_response(res)
        check_for_res_error(res_)
        return res_

    @overload
    def get(self, resource_path: str, *, raw: Literal[True]) -> HTTPResponse: ...
    @overload
    def get(self, resource_path: str, *, raw: Literal[False] = False) -> 'Union[dict, list[dict]]': ...

    def get(self, resource_path: str, *, raw: bool = False) -> 'Union[dict, list[dict], HTTPResponse]':
        """Send a 'GET' request to the configured Sedaro host.

        Returns a dictionary of the response data unless `raw` is set to `True`, in which case it returns a `urllib3`
        `HTTPResponse`.
        """
        return self.__request(resource_path, 'GET', raw=raw)

    def requests_lib_get(self, url: str) -> requests.Response:
        '''Get request using the requests library'''

        auth_header_name, auth_header_value = self.__sedaro._auth_header()

        kwargs = {
            'url': f'{self.__sedaro._api_host}{url}',
            'headers': {
                auth_header_name: auth_header_value,
                'Accept': 'application/json'  # Required for Sedaro firewall
            }
        }

        if self.__sedaro._csrf_token:
            kwargs['headers']['X-CSRFToken'] = self.__sedaro._csrf_token

        if self.__sedaro._proxy_url and self.__sedaro._proxy_url.startswith('http'):
            protocol = self.__sedaro._api_host.split('://')[0]
            kwargs['proxies'] = {protocol: self.__sedaro._proxy_url}
            kwargs['headers'] |= (self.__sedaro._proxy_headers or {})
            if not self.__sedaro._verify_ssl:
                kwargs['verify'] = False

        TIMEOUT_SECONDS = 180
        MAX_TIMEOUTS = 5
        timeout_iterations = 0
        while True:
            try:
                return requests.get(timeout=TIMEOUT_SECONDS, **kwargs)
            except (requests.exceptions.ConnectTimeout, requests.exceptions.ReadTimeout):
                if timeout_iterations < MAX_TIMEOUTS:
                    timeout_iterations += 1
                    continue
                raise

    @overload
    def post(self, resource_path: str, body: dict, *, raw: Literal[True]) -> HTTPResponse: ...
    @overload
    def post(self, resource_path: str, body: dict, *, raw: Literal[False] = False) -> dict: ...

    def post(self, resource_path: str, body: dict, *, raw: bool = False) -> 'dict | HTTPResponse':
        """Send a 'POST' request to the configured Sedaro host.

        Returns a dictionary of the response data unless `raw` is set to `True`, in which case it returns a `urllib3`
        `HTTPResponse`.
        """
        return self.__request(resource_path, 'POST', body, raw=raw)

    @overload
    def put(self, resource_path: str, body: dict, *, raw: Literal[True]) -> HTTPResponse: ...
    @overload
    def put(self, resource_path: str, body: dict, *, raw: Literal[False] = False) -> dict: ...

    def put(self, resource_path: str, body: dict, *, raw: bool = False) -> 'dict | HTTPResponse':
        """Send a 'PUT' request to the configured Sedaro host.

        Returns a dictionary of the response data unless `raw` is set to `True`, in which case it returns a `urllib3`
        `HTTPResponse`.
        """
        return self.__request(resource_path, 'PUT', body, raw=raw)

    @overload
    def patch(self, resource_path: str, body: dict, *, raw: Literal[True]) -> HTTPResponse: ...
    @overload
    def patch(self, resource_path: str, body: dict, *, raw: Literal[False] = False) -> dict: ...

    def patch(self, resource_path: str, body: dict, *, raw: bool = False) -> 'dict | HTTPResponse':
        """Send a 'PATCH' request to the configured Sedaro host.

        Returns a dictionary of the response data unless `raw` is set to `True`, in which case it returns a `urllib3`
        `HTTPResponse`.
        """
        return self.__request(resource_path, 'PATCH', body, raw=raw)

    @overload
    def delete(self, resource_path: str, *, raw: Literal[True]) -> HTTPResponse: ...
    @overload
    def delete(self, resource_path: str, *, raw: Literal[False] = False) -> dict: ...

    def delete(self, resource_path: str, *, raw: bool = False) -> 'dict | HTTPResponse':
        """Send a 'DELETE' request to the configured Sedaro host.

        Returns a dictionary of the response data unless `raw` is set to `True`, in which case it returns a `urllib3`
        `HTTPResponse`.
        """
        return self.__request(resource_path, 'DELETE', raw=raw)



import datetime
import math
import typing

import numpy as np
import pandas as pd
import pytz
from dateutil import parser

EPSILON = 1e-6
REF_DATETIME = datetime.datetime(2024, 3, 21, 0, 27, 23, tzinfo=datetime.timezone.utc)


def datetime_now_utc() -> datetime.datetime:
    '''Python version agnostic way to get the current UTC datetime.'''
    try:
        return datetime.datetime.now(datetime.timezone.utc)
    except ImportError:
        return datetime.datetime.utcnow()  # now deprecated


def datetime_to_mjd(dt: datetime.datetime) -> float:
    '''Convert a UTC datetime.datetime to a Modified Julian Date float.

    Args:
        dt: Datetime object with timezone information.

    Returns:
        float: Modified Julian Date.
    '''
    assert dt.tzinfo is not None, (
        "Datetime must have a timezone. You can define UTC datetimes via the `tzinfo=datetime.timezone.utc` argument."
    )
    return 60390.0190162037 + (dt - REF_DATETIME).total_seconds() / 86400


def mjd_to_datetime(mjd: float) -> datetime.datetime:
    '''Convert Modified Julian Date float to a UTC datetime.datetime.

    Args:
        mjd: Modified Julian Date float.

    Returns:
        datetime.datetime: Datetime object corresponding to the input MJD.
    '''
    return REF_DATETIME + datetime.timedelta(seconds=((mjd - 60390.0190162037)*86400))


def read_csv_time_series(file_path: str, time_column_header: str = 'time', **kwargs):
    '''Read a CSV file with a time column and return a pandas DataFrame with datetime objects for the time column.

    Args:
        file_path: Path to the CSV file.
        time_column_header: Header of the time column in the CSV file. Defaults to 'time'.
        **kwargs: Additional keyword arguments to pass to pandas.read_csv.

    Returns:
        pandas.DataFrame: DataFrame with datetime objects for the time column.
    '''
    data = pd.read_csv(file_path, **kwargs)
    datetimes = [parser.parse(t) for t in data[time_column_header]]
    for i, dt in enumerate(datetimes):
        if dt.tzinfo is None:
            datetimes[i] = pytz.utc.localize(dt)
    data[time_column_header] = datetimes
    return data


def read_excel_time_series(file_path: str, time_column_header: str = 'time', **kwargs):
    '''Read a Excel file with a time column and return a pandas DataFrame with datetime objects for the time column.

    Args:
        file_path: Path to the Excel file.
        time_column_header: Header of the time column in the CSV file. Defaults to 'time'.
        **kwargs: Additional keyword arguments to pass to pandas.read_csv.

    Returns:
        pandas.DataFrame: DataFrame with datetime objects for the time column.
    '''
    data = pd.read_excel(file_path, **kwargs)
    datetimes = []
    for dt in data[time_column_header]:
        if type(dt) is pd.Timestamp:
            datetimes.append(dt.to_pydatetime())
        elif type(dt) is datetime.datetime:
            datetimes.append(dt)
        else:
            datetimes.append(parser.parse(dt))
    for i, dt in enumerate(datetimes):
        if dt.tzinfo is None:
            datetimes[i] = pytz.utc.localize(dt)
    data[time_column_header] = datetimes
    return data


def search_time_series(time_dimension: typing.Union[np.ndarray, list], timestamp: typing.Union[float, datetime.datetime]) -> int:
    '''Proxy for [np.searchsorted](https://numpy.org/doc/stable/reference/generated/numpy.searchsorted.html).

    Args:
        time_dimension: List or array of time values in ascending order.
        timestamp: Time value to search for.

    Returns:
        int: Index of the rightmost element less than or equal to the target.

    Returns the index of the rightmost element less than or equal to the target.
    '''
    return np.searchsorted(time_dimension, timestamp, 'right')


def quaternion2attitude_mat(quaternion: np.ndarray) -> np.ndarray:
    '''Convert a quaternion to an attitude matrix which transforms vectors between frames specified by the quaternion.

    Args:
        quaternion: Quaternion with shape (4,). Scalar is the last element.

    Returns:
        ndarray: Rotation matrix with shape (3, 3).

    References:
        [1] F. Markley and John Crassidis. Fundamentals of Spacecraft Attitude Determination and Control. Section 2.9.3
    '''
    if len(quaternion) != 4:
        raise ValueError("Bad shape")

    rotation = np.zeros((3, 3))

    rotation[0, 0] = 1 - 2 * (quaternion[1] * quaternion[1] + quaternion[2] * quaternion[2])
    rotation[0, 1] = 2 * (quaternion[0] * quaternion[1] + quaternion[3] * quaternion[2])
    rotation[0, 2] = 2 * (quaternion[0] * quaternion[2] - quaternion[3] * quaternion[1])

    rotation[1, 0] = 2 * (quaternion[0] * quaternion[1] - quaternion[3] * quaternion[2])
    rotation[1, 1] = 1 - 2 * (quaternion[0] * quaternion[0] + quaternion[2] * quaternion[2])
    rotation[1, 2] = 2 * (quaternion[1] * quaternion[2] + quaternion[3] * quaternion[0])

    rotation[2, 0] = 2 * (quaternion[0] * quaternion[2] + quaternion[3] * quaternion[1])
    rotation[2, 1] = 2 * (quaternion[1] * quaternion[2] - quaternion[3] * quaternion[0])
    rotation[2, 2] = 1 - 2 * (quaternion[0] * quaternion[0] + quaternion[1] * quaternion[1])

    return rotation


def quaternion_rotate_frame(vectorIn: np.ndarray, quaternion: np.ndarray) -> np.ndarray:
    '''Rotate a vector into a frame specified by an attitude quaternion according to [1].

    Args:
        vectorIn: Input vector with shape (3,).
        quaternion: Quaternion with shape (4,). Scalar is the last element.

    Returns:
        ndarray: Rotated vector with shape (3,).

    References:
        [1] F. Markley and John Crassidis. Fundamentals of Spacecraft Attitude Determination and Control. Section 2.9.3
    '''
    return quaternion2attitude_mat(quaternion).dot(vectorIn)


def angle_between_quaternion(q1: np.ndarray, q2: np.ndarray) -> float:
    '''Calculate the angle between two quaternion attitudes.

    Args:
        q1: Quaternion with shape (4,).
        q2: Quaternion with shape (4,).

    Returns:
        float: Angle between the two quaternions, in radians.
    '''
    return 2 * np.arccos(min(abs(difference_quaternion(q1, q2)[-1]), 1))


def difference_quaternion(q1: np.ndarray, q2: np.ndarray) -> np.ndarray:
    '''Calculate quaternion describing rotation from q1 to q2.

    Args:
        q1: Origin quaternion with shape (4,).
        q2: Destination quaternion with shape (4,).

    Returns:
        ndarray: Relative quaternion with shape (4,).
    '''
    return quaternion_dot(q1, quaternion_conjugate(q2))


def quaternion2rotmat(quaternion: np.ndarray) -> np.ndarray:
    '''
    This function produces a rotation matrix corresponding to the ACTIVE rotation by the given quaternion. Don't use
    this for frame conversions; use quaternion2AttitudeMat instead.

    Args:
        quaternion: Quaternion with shape (4,). Scalar is the last element.

    Returns:
        ndarray: Rotation matrix with shape (3, 3).

    References:
        [1] F. Markley and John Crassidis. Fundamentals of Spacecraft Attitude Determination and Control.
    '''
    if len(quaternion) != 4:
        raise ValueError("Bad shape")

    rotation = np.zeros((3, 3))

    rotation[0, 0] = 1 - 2 * (quaternion[1] * quaternion[1] + quaternion[2] * quaternion[2])
    rotation[1, 0] = 2 * (quaternion[0] * quaternion[1] + quaternion[3] * quaternion[2])
    rotation[2, 0] = 2 * (quaternion[0] * quaternion[2] - quaternion[3] * quaternion[1])

    rotation[0, 1] = 2 * (quaternion[0] * quaternion[1] - quaternion[3] * quaternion[2])
    rotation[1, 1] = 1 - 2 * (quaternion[0] * quaternion[0] + quaternion[2] * quaternion[2])
    rotation[2, 1] = 2 * (quaternion[1] * quaternion[2] + quaternion[3] * quaternion[0])

    rotation[0, 2] = 2 * (quaternion[0] * quaternion[2] + quaternion[3] * quaternion[1])
    rotation[1, 2] = 2 * (quaternion[1] * quaternion[2] - quaternion[3] * quaternion[0])
    rotation[2, 2] = 1 - 2 * (quaternion[0] * quaternion[0] + quaternion[1] * quaternion[1])

    return rotation


def orthogonal_vector(vector: np.ndarray) -> np.ndarray:
    '''Produce an orthogonal vector.

    Args:
        vector: Input vector with shape (N,).

    Returns:
        ndarray: Output vector is orthogonal to input but otherwise arbitrarily oriented.
    '''
    cross_vector = np.array([0, 0, 1])
    normed_axis = vector / np.linalg.norm(vector)

    if abs(np.dot(normed_axis, cross_vector)) == 1:
        cross_vector = np.array([0, 1, 0])

    orthogonal_vector = np.cross(cross_vector, normed_axis)
    return orthogonal_vector / np.linalg.norm(orthogonal_vector)


def quaternion_dot(q1: np.ndarray, q2: np.ndarray) -> np.ndarray:
    '''Quaternion dot product.

    Preserves the order of active rotation matrix multiplication.

    Args:
        q1: Quaternion with shape (4,). Scalar is the last element.
        q2: Quaternion with shape (4,).

    Returns:
        ndarray: Quaternion with shape (4,).

    References:
        [1] F. Markley and John Crassidis. Fundamentals of Spacecraft Attitude Determination and Control.
    '''
    return np.array([q1[3] * q2[0] + q1[0] * q2[3] + q1[1] * q2[2] - q1[2] * q2[1],
                    q1[3] * q2[1] - q1[0] * q2[2] + q1[1] * q2[3] + q1[2] * q2[0],
                    q1[3] * q2[2] + q1[0] * q2[1] - q1[1] * q2[0] + q1[2] * q2[3],
                    q1[3] * q2[3] - q1[0] * q2[0] - q1[1] * q2[1] - q1[2] * q2[2]])


def random_orthogonal_rotation(
    vector: np.ndarray,
    angle_1sigma: float,
    random: typing.Optional[np.random.RandomState] = None,
) -> np.ndarray:
    '''Calculate a random rotation orthogonal to an input vector.

    Axis of rotation for the perturbation will be orthogonal to the input
    vector and uniformly distributed along the unit circle centered on
    and orthogonal to the input unit vector.

    Args:
        vector: Input vector with shape (N,).
        angle_1sigma: Standard deviation of angular perturbation.
        random: Pre-generated RandomState. Defaults to None.

    Returns:
        ndarray: Quaternion describing the random rotation with shape (4,).
    '''
    if random is None:
        random = np.random.RandomState()

    orth_vector = orthogonal_vector(vector)
    angle_error = angle_1sigma * random.randn()

    rotation_axis = quaternion2rotmat(
        euler_axis_angle2quaternion(vector, random.uniform(0, 2*np.pi))).dot(orth_vector)

    return euler_axis_angle2quaternion(rotation_axis, angle_error)


def euler_axis_angle2quaternion(axis: np.ndarray, angle: float) -> np.ndarray:
    '''Convert an axis-angle representation to a quaternion.

    Args:
        axis: Axis of rotation with shape (3,).
        angle: Angle of rotation in radians.

    Returns:
        ndarray: Quaternion with shape (4,).
    '''
    axis = axis / np.linalg.norm(axis)
    quaternion = np.array([axis[0] * np.sin(angle / 2),
                           axis[1] * np.sin(angle / 2),
                           axis[2] * np.sin(angle / 2),
                           np.cos(angle / 2)])
    return quaternion / np.linalg.norm(quaternion)


def vectors2angle(vector1: np.ndarray, vector2: np.ndarray) -> float:
    '''Find the acute angle between two vectors.

    Args:
        vector1: Vector with shape (3,).
        vector2: Vector with shape (3,).

    Returns:
        float: Angle between the two vectors in radians.
    '''
    cos_ang = np.dot(unit3(vector1), unit3(vector2))
    if abs(cos_ang) - 1 > 0:
        if cos_ang > 0:
            angle = 0.0
        else:
            angle = np.pi
    else:
        angle = math.acos(cos_ang)

    return angle


def eci_vector_to_body(vector_eci: np.ndarray, attitude_body_eci: np.ndarray) -> np.ndarray:
    '''Rotate a vector from ECI into the body frame given a `body_eci` quaternion.

    Args:
        vector_eci: Vector in ECI frame with shape (3,).
        attitude_body_eci: Quaternion with shape (4,). Scalar is the last element.

    Returns:
        ndarray: Rotated vector with shape (3,).
    '''
    rotation = quaternion2attitude_mat(attitude_body_eci)
    return rotation.dot(vector_eci)


def body_vector_to_eci(vector_eci: np.ndarray, attitude_body_eci: np.ndarray) -> np.ndarray:
    '''Rotate a vector from the body frame into ECI given a `body_eci` quaternion.

    Args:
        vector_eci: Vector in body frame with shape (3,).
        attitude_body_eci: Quaternion with shape (4,). Scalar is the last element.

    Returns:
        ndarray: Rotated vector with shape (3,).
    '''
    rotation = quaternion2attitude_mat(attitude_body_eci).T
    return rotation.dot(vector_eci)


def quaternion_conjugate(quaternion: np.ndarray) -> np.ndarray:
    '''Calculate the conjugate of a quaternion.

    Args:
        quaternion: Quaternion with shape (4,). Scalar is the last element.

    Returns:
        ndarray: Conjugate quaternion with shape (4,).
    '''
    quaternion_conj = np.array(-quaternion)
    quaternion_conj[3] = -quaternion_conj[3]

    return quaternion_conj


def rotmat2quaternion(rot_mat: np.ndarray) -> np.ndarray:
    '''Convert a rotation matrix to a quaternion.

    Args:
        rot_mat: Rotation matrix with shape (3, 3).

    Returns:
        ndarray: Quaternion with shape (4,). Scalar is the last element.
    '''
    tr = np.trace(rot_mat)
    idx = np.argmax([*np.diag(rot_mat), tr])
    quaternion = np.zeros(4)
    if idx != 3:
        i = idx
        j = (i + 1) % 3
        k = (j + 1) % 3

        quaternion[i] = 1 - tr + 2 * rot_mat[i, i]
        quaternion[j] = rot_mat[j, i] + rot_mat[i, j]
        quaternion[k] = rot_mat[k, i] + rot_mat[i, k]
        quaternion[3] = rot_mat[k, j] - rot_mat[j, k]
    else:
        quaternion[0] = rot_mat[2, 1] - rot_mat[1, 2]
        quaternion[1] = rot_mat[0, 2] - rot_mat[2, 0]
        quaternion[2] = rot_mat[1, 0] - rot_mat[0, 1]
        quaternion[3] = 1 + tr

    return quaternion / np.linalg.norm(quaternion)


def quaternions_to_rates(q1: np.ndarray, q2: np.ndarray, dt: float) -> np.ndarray:
    '''Finite difference calculation of angular velocity from two quaternions.

    Args:
        q1: Quaternion at time t with shape (4,). Scalar is the last element.
        q2: Quaternion at time t+dt with shape (4,). Scalar is the last element.
        dt: Time step.

    Returns:
        ndarray: Angular velocity vector with shape (3,).

    References:
        [1] https://mariogc.com/post/angular-velocity-quaternions/
    '''
    return (2 / dt) * np.array([
        q1[3]*q2[0] - q1[0]*q2[3] - q1[1]*q2[2] + q1[2]*q2[1],
        q1[3]*q2[1] + q1[0]*q2[2] - q1[1]*q2[3] - q1[2]*q2[0],
        q1[3]*q2[2] - q1[0]*q2[1] + q1[1]*q2[0] - q1[2]*q2[3]])


def invert3(m: np.ndarray) -> np.ndarray:
    '''Invert a 3x3 matrix.

    Faster for looping than np.linalg.inv for 3x3 matrices.

    Args:
        m: Matrix with shape (3, 3).

    Returns:
        ndarray: Inverted matrix with shape (3, 3).
    '''
    if m.shape != (3, 3):
        raise ValueError("Matrix must be 3x3.")
    inv = np.zeros((3, 3))
    inv[0, 0] = m[1, 1] * m[2, 2] - m[1, 2] * m[2, 1]
    inv[1, 0] = m[1, 2] * m[2, 0] - m[1, 0] * m[2, 2]
    inv[2, 0] = m[1, 0] * m[2, 1] - m[1, 1] * m[2, 0]

    inv[0, 1] = m[0, 2] * m[2, 1] - m[0, 1] * m[2, 2]
    inv[1, 1] = m[0, 0] * m[2, 2] - m[0, 2] * m[2, 0]
    inv[2, 1] = m[0, 1] * m[2, 0] - m[0, 0] * m[2, 1]

    inv[0, 2] = m[0, 1] * m[1, 2] - m[0, 2] * m[1, 1]
    inv[1, 2] = m[0, 2] * m[1, 0] - m[0, 0] * m[1, 2]
    inv[2, 2] = m[0, 0] * m[1, 1] - m[0, 1] * m[1, 0]

    inv /= inv[0, 0] * m[0, 0] + inv[0, 1] * m[1, 0] + inv[0, 2] * m[2, 0]

    return inv


def unit3(vec: np.ndarray) -> np.ndarray:
    '''Normalize a 3D vector.

    Args:
        vec: Input vector with shape (3,).

    Returns:
        ndarray: Normalized vector with shape (3,).
    '''
    if np.linalg.norm(vec) > EPSILON:
        return vec/np.linalg.norm(vec)
    raise ValueError("Vector cannot be the zero vector.")


def data_scale_unit(value: int, bytes: bool = False) -> tuple[float, str]:
    """
    Find the appropriate unit to scale the given bit value and return the scale along with the unit string.
    The input value can be divided by the returned scale to convert it to the returned unit.

    Args:
        value (float): The value to be scaled.
        bytes (bool): Whether to convert the value to bytes.

    Returns:
        tuple: A tuple containing the scale and the unit string.
    """
    units = ("", "k", "M", "G", "T", "P")
    if value == 0:
        return value, ""
    value = value if not bytes else value / 8
    exponent = min(max(math.floor(math.log10(abs(value)) // 3), 0), len(units) - 1)
    scale: int = (10 ** (exponent * 3))
    scale = scale if not bytes else scale * 8
    unit = units[exponent]
    return scale, unit



import json
import os
from pathlib import Path
from typing import TYPE_CHECKING, Dict, Generator, Union

from .series import SedaroSeries
from .utils import (ENGINE_EXPANSION, ENGINE_MAP, HFILL,
                    FromFileAndToFileAreDeprecated, get_column_names,
                    get_parquets, get_static_data, get_static_data_engines, hfill)

if TYPE_CHECKING:
    import dask.dataframe as dd


class SedaroBlockResult(FromFileAndToFileAreDeprecated):

    def __init__(
            self,
            structure,
            series: dict,
            stats: dict,
            column_index: dict,
            prefix: str,
            stats_to_plot: list = None,
            static_data: dict[str: dict] = None,
        ):
        '''Initialize a new block result.

        Block results are typically created through the .block method of
        SedaroAgentResult or the .from_file method of this class.
        '''
        if 'name' in structure:
            self.__name = structure['name']
        elif structure == 'root':
            self.__name = 'root'
        else:
            self.__name = '<Unnamed Block>'
        self.__structure = structure
        self.__series = series
        self.__stats = {}
        if stats is None:
            stats = {}
        for k in stats:
            self.__stats[k] = {kk: vv for kk, vv in stats[k].items() if kk.startswith(prefix)}
        self.__static_data = static_data if static_data is not None else {}
        self.__column_index = column_index
        self.__prefix = prefix
        self.__variables = []
        for stream in self.__column_index:
            for variable in self.__column_index[stream]:
                self.__variables.append(variable)
        self.__variables = sorted(list(self.__variables))
        self.stats_to_plot = stats_to_plot if stats_to_plot is not None else []

    def __getattr__(self, name: str) -> SedaroSeries:
        '''Get a particular variable by name.

        Typically invoked by calling .<VARIABLE_NAME> on an instance
        of SedaroBlockResult.
        '''
        prefix = f"{self.__prefix}{name}"
        for stream in self.__column_index:
            if name in self.__column_index[stream]:
                flattened_stats = {}
                for engine in self.__stats:
                    for k in self.__stats[engine]:
                        flattened_stats[k] = self.__stats[engine][k]
                return SedaroSeries(
                    name,
                    self.__series[stream],
                    flattened_stats,
                    self.__column_index[stream][name],
                    prefix,
                    self.stats_to_plot,
                    self.__name
                )
        else:
            raise ValueError(f'Variable "{name}" not found.')

    def __contains__(self, variable: str) -> bool:
        '''Check if this block contains a variable by name.'''
        return variable in self.variables

    def __iter__(self) -> Generator:
        '''Iterate through variables on this block.'''
        return (self.__getattr__(variable) for variable in self.variables)

    def __repr__(self) -> str:
        return f'SedaroBlockResult({self.name})'

    def subst_name(self, key):
        if self.__name == '<Unnamed Block>':
            return key
        else:
            return '.'.join([self.__name] + key.split('.')[1:])

    @property
    def name(self):
        return self.__name

    @property
    def dataframe(self) -> 'Dict[str, dd.DataFrame]':
        '''Get the raw Dask DataFrames for this block.'''
        # only include columns in this block, not columns in the dataframes that are for other blocks
        scoped_data = {}
        for stream in self.__series:
            column_names = get_column_names(self.__column_index[stream], self.__prefix)
            scoped_data[stream] = self.__series[stream][column_names]
        return scoped_data

    @property
    def modules(self):
        return self.__series.keys()

    @property
    def variables(self):
        return self.__variables

    def module_to_dataframe(self):
        raise NotImplementedError("")

    def variable(self, name: str) -> SedaroSeries:
        '''Query a particular variable by name.'''
        return self.__getattr__(name)

    def save(self, path: Union[str, Path]):
        '''Save the block result to a directory with the specified path.'''
        try:
            os.makedirs(path)
        except FileExistsError:
            if not (os.path.isdir(path) and any(os.scandir(path))):
                raise FileExistsError(
                    f"A file or non-empty directory already exists at {path}. Please specify a different path.")
        with open(f"{path}/class.json", "w") as fp:
            json.dump({'class': 'SedaroBlockResult'}, fp)
        os.mkdir(f"{path}/data")
        parquet_files = []
        for engine in self.__series:
            engine_parquet_path = f"{path}/data/{(pname := engine.replace('/', '.'))}"
            parquet_files.append(pname)
            df: 'dd' = self.__series[engine]
            df.to_parquet(engine_parquet_path)
        with open(f"{path}/meta.json", "w") as fp:
            json.dump({
                'structure': self.__structure,
                'column_index': self.__column_index,
                'prefix': self.__prefix,
                'parquet_files': parquet_files,
                'stats': self.__stats,
                'static': self.__static_data,
            }, fp)
        print(f"Block result saved to {path}.")

    @classmethod
    def load(cls, path: Union[str, Path]):
        '''Load a block result from the specified path.'''
        import dask.dataframe as dd

        with open(f"{path}/class.json", "r") as fp:
            archive_type = json.load(fp)['class']
            if archive_type != 'SedaroBlockResult':
                raise ValueError(f"Archive at {path} is a {archive_type}. Please use {archive_type}.load instead.")
        with open(f"{path}/meta.json", "r") as fp:
            meta = json.load(fp)
            structure = meta['structure']
            column_index = meta['column_index']
            prefix = meta['prefix']
            stats = meta['stats'] if 'stats' in meta else {}
            static_data = meta['static'] if 'static' in meta else {}
        engines = {}
        try:
            for agent in meta['parquet_files']:
                df = dd.read_parquet(f"{path}/data/{agent}")
                engines[agent.replace('.', '/')] = df
        except KeyError:
            for agent in get_parquets(f"{path}/data/"):
                df = dd.read_parquet(f"{path}/data/{agent}")
                engines[agent.replace('.', '/')] = df
        return cls(structure, engines, stats, column_index, prefix, static_data=static_data)

    def __has_stats(self, variable: str) -> bool:
        for engine in self.__stats:
            for series in self.__stats[engine]:
                if series.startswith(self.__prefix + variable):
                    return True
        return False

    def summarize(self) -> None:
        '''Summarize these results in the console.'''
        hfill()
        print("Sedaro Simulation Block Result Summary".center(HFILL))
        if self.name != '<Unnamed Block>':
            print(f"'{self.name}'".center(HFILL))
        hfill()

        print("🧩 Simulated Modules")
        for module in self.modules:
            print(f'    • {ENGINE_EXPANSION[ENGINE_MAP[module.split("/")[1]]]}')

        print("\n📋 Variables Available")
        for variable in self.variables:
            stats_marker = '\033[0;32m*\033[0;0m' if self.__has_stats(variable) else ' '
            print(f'    • {stats_marker} {variable}')
        hfill()
        print("❓ Query variables with .<VARIABLE_NAME>")
        if self.__stats:
            print("❓ Query statistics with .<VARIABLE_NAME>.stats('<STAT_NAME_1>', '<STAT_NAME_2>', ...)")
            print("📊 Variables with statistics available are marked with a \033[0;32m*\033[0;0m")
        if self.__static_data:
            hfill()
            engines_as_text = ", ".join(get_static_data_engines(self.__static_data))
            print(f"📦 Static data is available for this block in the following engine(s): {engines_as_text}.")
            print("📦 Query with .static_data('<ENGINE_NAME>') for that engine's static data on this block,")
            print("   or .static_data() to get this block's static data for all engines.")

    def value_at(self, mjd):
        return {variable: self.__getattr__(variable).value_at(mjd) for variable in self.variables}

    def stats(self, *args):
        if len(args) == 0:
            cleaned_stats = {}
            for agent in self.__stats:
                for key in self.__stats[agent]:
                    cleaned_stats[self.subst_name(key)] = self.__stats[agent][key]
            return cleaned_stats
        elif len(args) == 1:
            cleaned_stats = {}
            for agent in self.__stats:
                for key in self.__stats[agent]:
                    cleaned_stats[self.subst_name(key)] = self.__stats[agent][key][args[0]]
            return cleaned_stats
        else:
            dicts_to_return = []
            for arg in args:
                cleaned_stats = {}
                for agent in self.__stats:
                    for key in self.__stats[agent]:
                        cleaned_stats[self.subst_name(key)] = self.__stats[agent][key][arg]
                dicts_to_return.append(cleaned_stats)
            return tuple(dicts_to_return)

    def static_data(self, engine=None):
        return get_static_data(self.__static_data, "Block", engine=engine)



from typing import TYPE_CHECKING, ClassVar

from ..BaseModel import BaseModel
from ..BaseModelManager import BaseModelManager

if TYPE_CHECKING:
    from ..repository import Repository
    from ..workspace import Workspace


class Project(BaseModel):

    def delete(self):
        self._delete(query_params={'delete': 'true'})

    @property
    def repositories(self) -> 'list[Repository]':
        '''Get the repositories of the project.'''
        from ..repository import Repository
        return self._get_rel('repositories', Repository)

    @property
    def workspace(self) -> 'Workspace':
        '''Get the workspace of the project.'''
        from ..workspace import Workspace
        return self._get_rel('workspace', Workspace)


class ProjectManager(BaseModelManager[Project]):

    _BASE_PATH: 'ClassVar[str]' = '/projects'
    _MODEL: 'ClassVar' = Project



from __future__ import annotations

import json
from abc import ABC
from dataclasses import dataclass, field
from functools import cache
from typing import TYPE_CHECKING, TypeVar

from sedaro.settings import ID

if TYPE_CHECKING:
    from .BaseModelManager import BaseModelManager
    M = TypeVar('M', bound='BaseModel')


@dataclass
class BaseModel(ABC):
    _raw_data: 'dict'
    _model_manager: 'BaseModelManager' = field(repr=False)

    def __update_data(self, data: 'dict'):
        self._raw_data.update(data)

    def __str__(self):
        return f"{self.__class__.__name__}(id={self.id})"

    def __getattr__(self, name):
        try:
            return self._raw_data[name]
        except KeyError:
            raise AttributeError(f"{self.__class__.__name__} has no attribute '{name}'")

    def __hash__(self):
        return hash(json.dumps(self._raw_data, sort_keys=True))

    def __eq__(self, other):
        if not (isinstance(other, self.__class__) and (self.id == other.id)):
            return False

        if set(self._raw_data) != set(other._raw_data):
            return False

        return hash(self) == hash(other)

    @property
    def _sedaro(self):
        '''The SedaroApiClient instance.'''
        return self._model_manager._sedaro

    @property
    def id(self):
        '''The id of the model.'''
        return self._raw_data[ID]

    def update(self, **kwargs):
        '''Update the model with the given keyword arguments.'''
        res = self._sedaro.request.patch(
            self._model_manager._req_url(id=self.id), body=kwargs
        )
        self.__update_data(res)

    def refresh(self):
        '''Refresh the model data from the api.'''
        res_raw_data = self._model_manager.get(self.id)._raw_data
        self.__update_data(res_raw_data)
        self._get_rel.cache_clear()

    def delete(self):
        '''Delete the corresponding model.'''
        self._delete()

    def _delete(self, *, query_params: dict = None):
        self._sedaro.request.delete(
            self._model_manager._req_url(id=self.id, query_params=query_params)
        )

    @cache
    def _get_rel(self, field: 'str', base_model_type: 'type[M]') -> 'M' | list[M]:
        '''Get the related model(s) of the given field.'''
        from .BaseModelManager import BaseModelManager

        if (id_or_ids := self._raw_data.get(field)) in (None, []):
            return id_or_ids

        mod_man_cls = BaseModelManager._model_to_model_manager[base_model_type]
        mod_man = mod_man_cls(self._sedaro)

        if isinstance(id_or_ids, list):
            return [mod_man.get(id_) for id_ in id_or_ids]
        return mod_man.get(id_or_ids)



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Dict

from ..settings import DATA_SIDE, MANY_SIDE, ONE_SIDE, TYPE

if TYPE_CHECKING:
    from .branch import Branch


class Common(ABC):
    """Abstract class for common functionality between `Branch` and `Block` classes."""

    @property
    @abstractmethod
    def data(self) -> Dict:
        '''Access the dictionary corresponding to root or the `Block` this instance is associated with.'''
        pass

    @property
    @abstractmethod
    def _branch(self) -> 'Branch':
        '''Access the `Branch` this instance is associated with.'''
        pass

    @property
    @abstractmethod
    def _relationship_attrs(self) -> Dict:
        '''Access the relationship fields dictionary from the meta attributes corresponding to root or the `Block` this
        instance is associated with.'''
        pass

    def __getattr__(self, key: str) -> any:
        """Allows for dotting into the instance to access keys on the referenced Sedaro `Block` or root. Additionally,
        makes it so dotting into relationship fields returns `Block`s corresponding to the related Sedaro Blocks.

        Args:
            key (str): attribute being keyed into

        Raises:
            AttributeError: if the attribute doesn't exist

        Returns:
            any: the value of the corresponding attribute
        """
        if key not in self.data:
            raise make_attr_error(key, self.data[TYPE])
        val = self.data[key]

        if not self.is_rel_field(key):
            return val

        # If relationship is undefined, return None
        if val is None:
            return None

        side_type = self.get_rel_field_type(key)

        if side_type == MANY_SIDE:
            return [self._branch.block(id) for id in val]

        if side_type == DATA_SIDE:
            return {self._branch.block(id): data for id, data in val.items()}

        if side_type == ONE_SIDE:
            return self._branch.block(val)

        raise NotImplementedError(
            f'Unsupported relationship type on "{self.data[TYPE]}", attribute: "{key}".'
        )

    def is_rel_field(self, field: str) -> bool:
        """Checks if the given `field` is a relationship field on the associated Sedaro Block or root.

        Args:
            field (str): field to check

        Raises:
            TypeError: if the value of `field` is not a string

        Returns:
            bool: indicates if the given `field` is a relationship field or not
        """
        return field in self._relationship_attrs

    def get_rel_field_type(self, field: str) -> str:
        """Get the type of relationship of the field. Note: first call `is_rel_field` if you need to confirm `field` is
        a relationship field.

        Args:
            field (str): the field to get the relationship type for

        Raises:
            TypeError: if the value of `field` is not a string or not a relationship field
            KeyError: if the value of `field` does not correspond to any field on the associated Sedaro Block or root

        Returns:
            str: a string indicating the type of relationship field
        """
        if not self.is_rel_field(field):
            raise TypeError(
                f'The given field "{field}" is not a relationship field on "{self.data[TYPE]}".')

        return self._relationship_attrs[field][TYPE]


# ------ helper function and vars for this file only ------


def make_attr_error(field: str, name: str) -> str:
    return AttributeError(f'There is no "{field}" attribute on "{name}"')



from typing import TYPE_CHECKING, ClassVar

from ..BaseModel import BaseModel
from ..BaseModelManager import BaseModelManager

if TYPE_CHECKING:
    from ..project import Project
    from ..repository import Repository


class Workspace(BaseModel):

    @property
    def projects(self) -> 'list[Project]':
        '''Get the projects of the workspace.'''
        from ..project import Project
        return self._get_rel('projects', Project)

    @property
    def repositories(self) -> 'list[Repository]':
        '''Get the repositories of the workspace.'''
        from ..repository import Repository
        return self._get_rel('repositories', Repository)


class WorkspaceManager(BaseModelManager[Workspace]):

    _BASE_PATH: 'ClassVar[str]' = '/workspaces'
    _MODEL: 'ClassVar' = Workspace



from abc import ABC
from dataclasses import dataclass
from typing import TYPE_CHECKING, ClassVar, Generic, TypeVar, overload
from urllib.parse import urlencode

if TYPE_CHECKING:
    from ..sedaro_api_client import SedaroApiClient
    from .BaseModel import BaseModel

M = TypeVar('M', bound='BaseModel')


@dataclass
class BaseModelManager(ABC, Generic[M]):

    _sedaro: 'SedaroApiClient'
    '''The SedaroApiClient instance'''

    _MODEL: 'ClassVar[type[M]]'
    '''The model class that this manager manages'''
    _BASE_PATH: 'ClassVar[str]'
    '''The base path for the model'''

    _model_to_model_manager: 'ClassVar[dict[type[BaseModel], type[BaseModelManager]]]' = {}
    '''Mapping of model to model manager, automatically set by __init_subclass__'''

    def __init_subclass__(cls):
        if cls._MODEL in cls._model_to_model_manager:
            raise ValueError(
                f'Attempting to have multiple managers ({cls} and {cls._model_to_model_manager[cls._MODEL]}) for the same model: {cls._MODEL}'
            )
        cls._model_to_model_manager[cls._MODEL] = cls
        return super().__init_subclass__()

    @overload
    def get(self) -> 'list[M]': ...
    @overload
    def get(self, id: str) -> 'M': ...

    def get(self, id: 'str' = None, /) -> 'list[M] | M':
        '''Get all accessible models or a single model by id'''
        if id is None:
            return [
                self._MODEL(w, self) for w in
                self._sedaro.request.get(self._req_url())
            ]

        return self._MODEL(
            self._sedaro.request.get(self._req_url(id=id)),
            self
        )

    def get_all(self) -> 'list[M]':
        '''Get all accessible models (same as `get()` with no arguments)'''
        return self.get()

    def create(self, **body) -> 'M':
        '''Create a new model with the given keyword arguments'''
        return self._create(**body)

    def _create(self, *, _id_in_url='', **body) -> 'M':
        w = self._sedaro.request.post(self._req_url(id=_id_in_url), body)
        return self._MODEL(w, self)

    __REQ_URL_NO_EXPAND = {'expand': {}}

    def _req_url(self, *, id: str = None, query_params: 'dict' = None) -> 'str':
        '''Return the request url for the given id and query parameters'''
        if query_params is None:
            query_str = f'?{urlencode(self.__REQ_URL_NO_EXPAND)}'
        else:
            query_str = f'?{urlencode(query_params | self.__REQ_URL_NO_EXPAND)}'

        id_str = f'/{id}' if id not in (None, '') else ''
        return f'{self._BASE_PATH}{id_str}{query_str}'



from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, List, Union

from ...exceptions import NoBlockFoundError
from ...settings import BLOCKS, CRUD, DATA_SIDE, ID, INDEX, MANY_SIDE, ONE_SIDE, TYPE
from ...utils import enforce_id_in_branch
from .block import Block

if TYPE_CHECKING:
    from ..branch import Branch


@dataclass
class BlockType:
    '''Class for getting `Block`s associated with Sedaro Blocks of this class type'''
    type: str
    '''Name of the Sedaro Block class this `BlockType` is set up to interact with'''
    _branch: 'Branch'
    '''The `Branch` this `BlockType` is connected to'''

    def __str__(self) -> str:
        return f'{self.__class__.__name__}({self.type}, branch={self._branch.id})'

    def __repr__(self):
        return self.__str__()

    def __hash__(self):
        return hash(f'{self.__class__.__name__}-{self.type}-{self._branch.id}')

    def create(self, **fields) -> Block:
        """Creates a Sedaro Block of the given type in the corresponding Branch. Note that if 'id' or 'type' are passed
        as kwargs, they will be ignored.

        Args:
            **fields (any): required and optional fields on the corresponding Sedaro Block.

        Raises:
            SedaroApiException: if there is an error in the response

        Returns:
            Block: a client to interact with the created Sedaro Block
        """

        for kwarg in [ID, TYPE]:
            fields.pop(kwarg, None)

        res = self._branch.crud(blocks=[{**fields, **{TYPE: self.type}}])
        block_id = res[CRUD][BLOCKS][-1]
        return Block(block_id, self)

    def get(self, id: Union[str, int]) -> Block:
        """Gets a `Block` of the desired type associated with the Sedaro Block of the given `id`.

        Args:
            id (Union[str, int]): `id` of the desired Sedaro Block

        Raises:
            KeyError: if no corresponding Block exists of the desired type

        Returns:
            Block: a client to interact with the corresponding Sedaro Block
        """
        enforce_id_in_branch(self._branch, id)

        # in addition to checks in ^^^ also make sure is the correct type for this BlockType
        if id not in self.get_all_ids():
            raise KeyError(
                f'There is no "{self.type}" with id "{id}" in this Branch.')

        return Block(id, self)

    def get_all_ids(self) -> List[str]:
        """Gets a `list` of `id`s corresponding to all Sedaro Blocks of the given type in this Branch. If there are no
        corresponding Blocks, returns an empty `list`.

        Returns:
            List[str]: list of `id`s
        """
        index = self._branch.data[INDEX]

        res = []

        def recurse_get_block_dicts(block_type):
            for type_or_id in index[block_type]:
                if type_or_id in index:
                    recurse_get_block_dicts(type_or_id)
                else:
                    res.append(type_or_id)

        recurse_get_block_dicts(self.type)

        # maintain order & filter out duplicates
        return list(dict.fromkeys(res))

    def get_all(self) -> List['Block']:
        """Gets a `list` of all `Block` instances corresponding to all Sedaro Blocks of the given type in this
        Branch. If there are no corresponding Blocks, returns an empty `list`.

        Returns:
            List['Block']: a list of `Block` instances corresponding to Sedaro Blocks in this Branch
        """
        return [Block(id, self) for id in self.get_all_ids()]

    def get_where(self, **fields) -> List['Block']:
        """
        Gets a filtered `list` of all `Block` instances corresponding to all Sedaro Blocks of the given type in
        this Branch. Blocks are filtered by property/values passed as kwargs. If there are no corresponding Blocks,
        returns an empty `list`.

        If querying via a relationship field, values can be passed as `Block` instances or `id`s. If querying on a Many
        side relationship specifically, the value is compared independent of order.

        **fields:
            any: keys to check for given values on the Sedaro Blocks

        Returns:
            List['Block']: a filtered list of `Block` instances corresponding to Sedaro Blocks in this\
                Branch
        """
        return [
            block for block in self.get_all()
            if all(self.__get_where_check_equality(block, k, v) for k, v in fields.items())
        ]

    def __get_where_check_equality(self, block_to_check: 'Block', attr: 'str', expected_val: Any) -> 'bool':
        '''Checks if the given `block_to_check` has the expected value for the given attribute `attr`. Turn all values
        into `id`s if they are `Block` instances in relationship fields.'''
        MISSING = object()
        # get `actual_val` from `data` rather than the Block itself to avoid rel values turning into Blocks
        if (actual_val := block_to_check.data.get(attr, MISSING)) is MISSING:
            return False

        if not block_to_check.is_rel_field(attr):
            return actual_val == expected_val

        side_type = block_to_check.get_rel_field_type(attr)

        if side_type == MANY_SIDE:
            if not isinstance(expected_val, list):
                return False
            expected_val = [
                v.id if isinstance(v, Block) else v
                for v in expected_val
            ]
            try:
                return set(expected_val) == set(actual_val)
            except:
                return False

        if side_type == DATA_SIDE:
            if not isinstance(expected_val, dict):
                return False
            expected_val = {
                k.id if isinstance(k, Block) else k: v
                for k, v in expected_val.items()
            }
            return expected_val == actual_val

        if side_type == ONE_SIDE:
            expected_val = expected_val.id if isinstance(expected_val, Block) else expected_val
            return actual_val == expected_val

        raise NotImplementedError(
            f'Unsupported relationship type on "{self.type}", attribute: "{attr}".'
        )

    def get_first(self):
        """Returns a `Block` associated with the least recently added (lowest `id`) Sedaro Block of the desired
        type.

        Raises:
            NoBlockFoundError: if no Blocks of the desired type exist in this Branch

        Returns:
            Block: a client to interact with the corresponding Sedaro Block
        """
        all_ids = self.get_all_ids()
        if not len(all_ids):
            raise NoBlockFoundError(
                f'No "{self.type}" Blocks exist in this Branch.'
            )
        return Block(sorted(all_ids)[0], self)

    def get_last(self):
        """Returns a `Block` associated with the most recently added (highest `id`) Sedaro Block of the desired
        type.

        Raises:
            NoBlockFoundError: if no Blocks of the desired type exist in this Branch

        Returns:
            Block: a client to interact with the corresponding Sedaro Block
        """
        all_ids = self.get_all_ids()
        if not len(all_ids):
            raise NoBlockFoundError(
                f'No "{self.type}" Blocks exist in this Branch.'
            )
        return Block(sorted(all_ids)[-1], self)



# coding: utf-8

"""
    Sedaro API

     Allows for consumption of Sedaro services. Read more about Sedaro at [docs.sedaro.com](https://docs.sedaro.com).  ### Clients  **Python:** [sedaro](https://pypi.org/project/sedaro/) - This package provides additional functionality on top of the auto-generated OpenAPI client. See the package docs for more information.  ### API Key  To access the Sedaro service via this API, you will need an API key.  You can generate an API key for your account in the Sedaro [Management Console](https://satellite.sedaro.com/account). Once complete, pass the API key in all requests via the `X_API_KEY` HTTP header.  *API keys grant full access to your account and should never be shared. If you think your API key has been compromised, you can revoke it in the [Management Console](https://satellite.sedaro.com/account).*  ### Jupyter Notebooks  For additional examples of how to use this API for modeling and simulation, see our [Mod-Sim Notebooks](https://github.com/sedaro/modsim-notebooks).  ### Community, Support, Discussion  If you have any issues or suggestions, please reach out:  1. Join the Sedaro Community [Slack](https://join.slack.com/t/sedaro-community/shared_invite/zt-1jps4i711-mXy88AZQ9AV7YcEXr8x7Ow) 2. Email us at support@sedarotech.com  ### Known Issues  - Error responses are more specific than what is shown throughout the documentation.  A 4xx or 5xx error will be returned in all error cases.  Only a `200` status indicates success.  See a given error response for additional details.   # noqa: E501

    The version of the OpenAPI document: 4.11.36
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from sedaro_base_client import schemas  # noqa: F401


class ServicesModelSpecModelsStudyJobStatuses(
    schemas.EnumBase,
    schemas.StrSchema
):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    An enumeration.
    """


    class MetaOapg:
        enum_value_to_name = {
            "PENDING": "PENDING",
            "RUNNING": "RUNNING",
            "PAUSED": "PAUSED",
            "TERMINATED": "TERMINATED",
            "FAILED": "FAILED",
            "SUCCEEDED": "SUCCEEDED",
            "ERROR": "ERROR",
        }
    
    @schemas.classproperty
    def PENDING(cls):
        return cls("PENDING")
    
    @schemas.classproperty
    def RUNNING(cls):
        return cls("RUNNING")
    
    @schemas.classproperty
    def PAUSED(cls):
        return cls("PAUSED")
    
    @schemas.classproperty
    def TERMINATED(cls):
        return cls("TERMINATED")
    
    @schemas.classproperty
    def FAILED(cls):
        return cls("FAILED")
    
    @schemas.classproperty
    def SUCCEEDED(cls):
        return cls("SUCCEEDED")
    
    @schemas.classproperty
    def ERROR(cls):
        return cls("ERROR")



from typing import TYPE_CHECKING

from sedaro_base_client.paths.models_branches_branch_id.get import SchemaFor200ResponseBodyApplicationJson

from ... import modsim as ms
from ...settings import SCENARIO
from ..blocks import BlockType
from ..branch import Branch

if TYPE_CHECKING:
    from ...sedaro_api_client import SedaroApiClient


class ScenarioBranch(Branch):

    def __init__(self, body: SchemaFor200ResponseBodyApplicationJson, sedaro: 'SedaroApiClient'):
        super().__init__(body, sedaro)
        if (type_ := self.data['type']) != SCENARIO:
            raise TypeError(
                f'Branch must be of type "{SCENARIO}" not "{type_}"')

    @property
    def simulation(self):
        """A `Simulation` instance to interact with the simulation connected to this scenario branch.

        Returns:
            Simulation: a `Simulation`
        """
        from .sim_client import Simulation
        return Simulation(self._sedaro, self)

    @property
    def study(self):
        """A `Study` instance to interact with a study connected to this scenario branch.

        Returns:
            Study: a `Study`
        """
        from .study_client import Study
        return Study(self._sedaro, self)

    def delete_all_external_state_blocks(self):
        """Delete all ExternalState blocks in the scenario branch."""
        if existing_externals := self.ExternalState.get_all_ids():
            self.crud(delete=existing_externals)

    def set_start_time_to_now(self, elapsed_time_seconds):
        """Set the start time of the scenario to the current time and the stop time to the current time plus the passed elapsed time in seconds."""
        start_time_mjd = ms.datetime_to_mjd(ms.datetime_now_utc())
        return self.clockConfig.update(
            startTime=start_time_mjd,
            stopTime=start_time_mjd+(elapsed_time_seconds/86400)
        )

    # ==============================================================================================================
    # For intellisense
    # ==============================================================================================================

    # next line used to know where to start auto edit
    # $AUTO_EDIT_START$

    AgentGroup: BlockType
    """A Sedaro `Block` class on a `ScenarioBranch`"""
    ClockConfig: BlockType
    """A Sedaro `Block` class on a `ScenarioBranch`"""
    Menu: BlockType
    """A Sedaro `Block` class on a `ScenarioBranch`"""
    MenuItem: BlockType
    """A Sedaro `Block` class on a `ScenarioBranch`"""
    PerRoundExternalState: BlockType
    """A Sedaro `Block` class on a `ScenarioBranch`"""
    SpontaneousExternalState: BlockType
    """A Sedaro `Block` class on a `ScenarioBranch`"""
    WaypointPathWithDuration: BlockType
    """A Sedaro `Block` class on a `ScenarioBranch`"""
    WaypointPathWithSpeed: BlockType
    """A Sedaro `Block` class on a `ScenarioBranch`"""
    WaypointPathWithTimestamps: BlockType
    """A Sedaro `Block` class on a `ScenarioBranch`"""
    WidgetSpec: BlockType
    """A Sedaro `Block` class on a `ScenarioBranch`"""
    PeripheralGroundArea: BlockType
    """A Sedaro `Block` class on a `ScenarioBranch`"""
    SpkEphemeris: BlockType
    """A Sedaro `Block` class on a `ScenarioBranch`"""
    PeripheralSpacePoint: BlockType
    """A Sedaro `Block` class on a `ScenarioBranch`"""
    PropagatedOrbitKinematics: BlockType
    """A Sedaro `Block` class on a `ScenarioBranch`"""
    EcefStationaryKinematics: BlockType
    """A Sedaro `Block` class on a `ScenarioBranch`"""
    PeripheralGroundPoint: BlockType
    """A Sedaro `Block` class on a `ScenarioBranch`"""
    OverrideSet: BlockType
    """A Sedaro `Block` class on a `ScenarioBranch`"""
    StkEphemeris: BlockType
    """A Sedaro `Block` class on a `ScenarioBranch`"""
    PeripheralAgent: BlockType
    """A Sedaro `Block` class on a `ScenarioBranch`"""
    PeripheralCelestialPoint: BlockType
    """A Sedaro `Block` class on a `ScenarioBranch`"""
    TemplatedAgent: BlockType
    """A Sedaro `Block` class on a `ScenarioBranch`"""
    AuxiliaryService: BlockType
    """A Sedaro `Block` class on a `ScenarioBranch`"""
    DynamicWaypointPath: BlockType
    """A Sedaro `Block` class on a `ScenarioBranch`"""



from sedaro_base_client.paths.models_branches_branch_id.get import ApiForget
from sedaro_base_client.paths.models_branches_branch_id.post import ApiForpost
from sedaro_base_client.paths.models_branches_branch_id.delete import ApiFordelete
from sedaro_base_client.paths.models_branches_branch_id.patch import ApiForpatch


class ModelsBranchesBranchId(
    ApiForget,
    ApiForpost,
    ApiFordelete,
    ApiForpatch,
):
    pass



from typing import Tuple

StreamData = dict[str, list]
Stream = Tuple[list, dict]
Streams = dict[str, Stream]

def concat_stream_data(main: StreamData, other: StreamData, len_main: int, len_other: int):
    assert type(main) == dict and type(other) == dict
    for k in other:
        if k not in main:
            main[k] = [None for _ in range(len_main)]
        main[k].extend(other[k])
    for k in main:
        if k not in other:
            main[k].extend([None for _ in range(len_other)])


def concat_stream(main: Stream, other: Stream, stream_id: str):
    len_main = len(main[0])
    len_other = len(other[0])
    main[0].extend(other[0])
    stream_id_short = stream_id.split('/')[0]
    concat_stream_data(main[1][stream_id_short], other[1][stream_id_short], len_main, len_other)


def concat_results(main: Streams, other: Streams):
    for stream in other:
        if stream not in main:
            main[stream] = other[stream]
        else:  # concat stream parts
            concat_stream(main[stream], other[stream], stream)


def update_metadata(main, other):
    for k in other['counts']:
        if k not in main['counts']:
            main['counts'][k] = 0
        main['counts'][k] += other['counts'][k]


def set_numeric_as_list(d):
    if isinstance(d, dict):
        if all(key.isdigit() for key in d.keys()):  # Check if all keys are array indexes in string form
            return [set_numeric_as_list(d[key]) for key in sorted(d.keys(), key=int)]
        else:
            return {k: set_numeric_as_list(v) for k, v in d.items()}
    return d


def __set_nested(results):
    nested = {}
    for k in sorted(list(results.keys())):
        v = results[k]
        try:
            ptr = nested
            tokens = k.split('.')
            for token in tokens[:-1]:
                if token not in ptr:
                    ptr[token] = {}
                ptr = ptr[token]
            ptr[tokens[-1]] = v
        except TypeError:
            ptr = nested
            for token in tokens[:-1]:
                if type(ptr[token]) == list:
                    del ptr[token]
                    break
                else:
                    ptr = ptr[token]
            ptr = nested
            for token in tokens[:-1]:
                if token not in ptr:
                    ptr[token] = {}
                ptr = ptr[token]
            ptr[tokens[-1]] = v
    return nested


# TODO: edge case where one page has all nones for a SV, then the next page has a bunch of vectors for it
def set_nested(results):
    nested = {}
    for k in results:
        kspl = k.split('/')[0]
        nested[k] = (results[k][0], {kspl: set_numeric_as_list(__set_nested(results[k][1][kspl]))})
    return nested



import datetime as dt
import json
import os
from pathlib import Path
import time
from typing import TYPE_CHECKING, Dict, List, Union

from ..settings import STATUS, SUCCEEDED
from ..branches.scenario_branch.utils import _get_stats_for_sim_id
from .agent import SedaroAgentResult
from .utils import (HFILL, STATUS_ICON_MAP, FromFileAndToFileAreDeprecated,
                    _block_type_in_supers, _get_agent_id_name_map,
                    _restructure_data, get_parquets, hfill)

if TYPE_CHECKING:
    import dask.dataframe as dd
    from ..sedaro_api_client import SedaroApiClient


class SimulationResult(FromFileAndToFileAreDeprecated):

    def __init__(self, simulation: dict, data: dict, _sedaro: 'SedaroApiClient' = None, stats_fetched: bool = False):
        '''Initialize a new Simulation Result using methods on the `simulation` property of a `ScenarioBranch`.

        See the `from_file` class method on this class for alternate initialization.
        '''
        self.__simulation = {
            'id': simulation.get('id', None),
            'branch': simulation['branch'],
            'dateCreated': simulation['dateCreated'],
            'dateModified': simulation['dateModified'],
            STATUS: str(simulation[STATUS]),
        }
        self.__sedaro = _sedaro
        self.__branch = simulation['branch']
        self.__data = data
        self.__meta: dict = data['meta']
        self.__stats_fetched = ('stats_fetched' in data and data['meta']['stats_fetched']) or ('stats' in data and data['stats'])
        self.__stats = data['stats'] if 'stats' in data else {}
        self.__static_data: dict = data['static'] if 'static' in data else {}
        self.stats_to_plot = []
        raw_series = data['series']
        agent_id_name_map = _get_agent_id_name_map(self.__meta)
        self.__agent_ids, self.__block_structures, self.__column_index = _restructure_data(
            raw_series, agent_id_name_map, self.__meta)

    def __repr__(self) -> str:
        return f'SedaroSimulationResult(branch={self.__branch}, status={self.status})'

    def fetch_stats(self, wait=False):
        '''Fetch the stats for this SimulationResult.

        If `wait` is True, this method will block until the stats are ready.
        '''
        if self.__stats_fetched:
            print("Stats already fetched.")
            return
        if self.__sedaro is None:
            raise Exception("Fetching summary stats after loading from save is not currently supported.")
        result, success = _get_stats_for_sim_id(self.__sedaro, self.__meta['id'])
        if success:
            self.__stats = result
            self.__stats_fetched = True
        else:
            if wait:
                POLLING_INTERVAL = 2.0
                while not success:
                    time.sleep(POLLING_INTERVAL)
                    result, success = _get_stats_for_sim_id(self.__sedaro, self.__meta['id'])
                self.__stats = result
                self.__stats_fetched = True
            else:
                raise Exception("Failed to fetch summary stats for simulation. Stats are not yet ready.")
        print("Stats fetched.")

    @property
    def stats_fetched(self) -> bool:
        return self.__stats_fetched

    @property
    def job_id(self):
        return self.__simulation['id']

    @property
    def data_array_id(self):
        return self.__meta.get('id', None)

    @property
    def templated_agents(self) -> List[str]:
        return tuple([
            entry['name'] for _, entry
            in self.__meta['structure']['scenario']['blocks'].items()
            if _block_type_in_supers(entry['type'], self.__meta['structure']['scenario']['_supers'], super_type='TemplatedAgent')
        ])

    @property
    def peripheral_agents(self) -> List[str]:
        return tuple([
            entry['name'] for id_, entry
            in self.__meta['structure']['scenario']['blocks'].items()
            if _block_type_in_supers(entry['type'], self.__meta['structure']['scenario']['_supers'], super_type='PeripheralAgent') and id_ in self.__agent_ids
        ])

    @property
    def dataframe(self) -> 'Dict[str, dd.DataFrame]':
        '''Get the raw Dask DataFrames for this SimulationResult.'''
        return self.__data['series']

    @property
    def status(self) -> str:
        return str(self.__simulation[STATUS])

    @property
    def start_time(self) -> dt.datetime:
        return dt.datetime.strptime(self.__simulation['dateCreated'], "%Y-%m-%dT%H:%M:%S.%fZ")

    @property
    def end_time(self) -> dt.datetime:
        return dt.datetime.strptime(self.__simulation['dateModified'], "%Y-%m-%dT%H:%M:%S.%fZ")

    @property
    def run_time(self) -> float:
        return (self.end_time - self.start_time).total_seconds()

    @property
    def success(self) -> bool:
        return str(self.__simulation[STATUS]) == SUCCEEDED

    def __assert_success(self) -> None:
        if not self.success:
            raise ValueError(
                'This operation cannot be completed because the simulation hasn\'t finished or failed early.')

    def __agent_id_from_name(self, name: str) -> str:
        for id_, entry in self.__meta['structure']['scenario']['blocks'].items():
            if name == entry.get('name') and _block_type_in_supers(entry['type'], self.__meta['structure']['scenario']['_supers']):
                if id_ in self.__agent_ids:
                    return id_
        else:
            raise ValueError(f"Agent {name} not found in data set.")

    def agent(self, id_or_name: str) -> SedaroAgentResult:
        '''Query results for a particular agent by name or ID. In the event that an agent's name is the same as another agent's ID, the agent with the ID that matches the given string will be used.'''
        if id_or_name in self.__agent_ids:
            agent_id, name = id_or_name, self.__agent_ids[id_or_name]
        else:
            try:
                agent_id, name = self.__agent_id_from_name(id_or_name), id_or_name
            except ValueError:
                raise ValueError(f"Agent with `id` or `name` '{id_or_name}' not found in data set. If an expected agent is missing, the simulation may have terminated early.")
        agent_dataframes = {}
        for stream_id in self.__data['series']:
            if stream_id.startswith(agent_id):
                agent_dataframes[stream_id] = self.__data['series'][stream_id]
        initial_agent_models = self.__meta['structure']['agents']
        initial_state = initial_agent_models[agent_id] if agent_id in initial_agent_models else None
        filtered_stats = {k: v for k, v in self.__stats.items() if k.startswith(agent_id)}
        filtered_static_data = {k: v for k, v in self.__static_data.items() if k.startswith(agent_id)}
        return SedaroAgentResult(
            name,
            self.__block_structures[agent_id],
            agent_dataframes,
            self.__column_index[agent_id],
            initial_state=initial_state,
            stats=filtered_stats,
            stats_to_plot=self.stats_to_plot,
            static_data=filtered_static_data,
        )

    def save(self, path: Union[str, Path]):
        '''Save the simulation result to a directory with the specified path.'''
        try:
            os.makedirs(path)
        except FileExistsError:
            if not (os.path.isdir(path) and any(os.scandir(path))):
                raise FileExistsError(
                    f"A file or non-empty directory already exists at {path}. Please specify a different path.")
        with open(f"{path}/class.json", "w") as fp:
            json.dump({'class': 'SimulationResult'}, fp)
        os.mkdir(f"{path}/data")
        parquet_files = []
        for agent in self.__data['series']:
            agent_parquet_path = f"{path}/data/{(pname := agent.replace('/', '.'))}"
            parquet_files.append(pname)
            df: 'dd' = self.__data['series'][agent]
            df.to_parquet(agent_parquet_path)
        with open(f"{path}/meta.json", "w") as fp:
            json.dump({'meta': self.__data['meta'], 'simulation': self.__simulation,
                      'stats': self.__stats, 'static': self.__static_data, 'parquet_files': parquet_files}, fp)
        print(f"Simulation result saved to {path}.")

    @classmethod
    def load(cls, path: Union[str, Path]):
        '''Load a simulation result from the specified path.'''
        import dask.dataframe as dd
        with open(f"{path}/class.json", "r") as fp:
            archive_type = json.load(fp)['class']
            if archive_type != 'SimulationResult':
                raise ValueError(f"Archive at {path} is a {archive_type}. Please use {archive_type}.load instead.")
        data = {}
        with open(f"{path}/meta.json", "r") as fp:
            contents = json.load(fp)
            simulation = contents['simulation']
            data['meta'] = contents['meta']
            data['stats'] = contents['stats'] if 'stats' in contents else {}
            data['static'] = contents['static'] if 'static' in contents else {}
        data['series'] = {}
        try:
            for agent in contents['parquet_files']:
                df = dd.read_parquet(f"{path}/data/{agent}")
                data['series'][agent.replace('.', '/')] = df
        except KeyError:
            for agent in get_parquets(f"{path}/data/"):
                df = dd.read_parquet(f"{path}/data/{agent}")
                data['series'][agent.replace('.', '/')] = df
        return cls(simulation, data)

    def summarize(self) -> None:
        '''Summarize these results in the console.'''
        hfill()
        print(f'Sedaro Simulation Result Summary'.center(HFILL))
        hfill()
        print(
            f'{STATUS_ICON_MAP[self.status]} Simulation {self.status.lower()} after {self.run_time:.1f}s')

        agents = self.templated_agents
        if len(agents) > 0:
            print('\n🛰️ Templated Agents ')
            for entry in agents:
                print(f'    • {entry}')

        agents = self.peripheral_agents
        if len(agents) > 0:
            print('\n📡 Peripheral Agents ')
            for entry in agents:
                print(f'    • {entry}')

        hfill()
        print("❓ Query agent results with .agent(<NAME>)")



# import warnings

# from tqdm import TqdmWarning, tqdm

# imprecision of float math sometimes forces Tqdm to clamp a number to a range
# this suppresses the warning that prints each time that happens
# warnings.filterwarnings('ignore', category=TqdmWarning)

class NonTqdmProgressBar:
    def __init__(self, *args, **kwargs):
        print("Downloading...")
        self.n = 0
        pass

    def update(self, *args, **kwargs):
        pass

    def refresh(self):
        pass

    def close(self):
        print("...download complete!")


class ProgressBar:
    def __init__(self, start, stop, num_streams, desc):
        # self.bar = tqdm(range(num_streams), desc=desc,  bar_format='{l_bar}{bar}[{elapsed}<{remaining}]')
        self.bar = NonTqdmProgressBar()
        self.num_streams = num_streams
        self.start = start
        self.stop = stop
        self.prev = {}

    def update(self, stream_id, new):
        # used for the download progress bar. Updates the progress bar based on the time range
        # fetched for a stream by the latest page.
        if stream_id not in self.prev:
            self.prev[stream_id] = 0
        if new > self.stop:
            new = self.stop
        incr = ((new - self.start) / (self.stop - self.start)) - self.prev[stream_id]
        # sometimes the elapsed time goes over the sim stop time
        # this would cause the progress bar to go over 100%, so we clamp it
        self.bar.update(min(incr, self.num_streams - self.bar.n))
        self.bar.refresh()
        self.prev[stream_id] = incr

    def complete(self):
        # set the progress bar to EXACTLY 100% and close it
        self.bar.update(self.num_streams - self.bar.n)  # see https://github.com/tqdm/tqdm/issues/1264
        self.bar.refresh()
        self.bar.close()


class StreamManager:
    def __init__(self, download_bar):
        self.dataframe = None
        self.keys = set()
        self.download_bar = download_bar

    def ingest_core_data(self, stream_id, core_data):
        import dask.dataframe as dd
        if self.dataframe is None:
            self.dataframe = dd.from_dict(core_data, npartitions=1)
        else:
            self.dataframe = dd.concat([self.dataframe, dd.from_dict(core_data, npartitions=1)], axis=0)
        self.keys.update(core_data.keys())
        self.download_bar.update(stream_id, core_data['time'][-1])

    def ingest(self, stream_id, stream_data):
        core_data = stream_data[1][stream_id.split('/')[0]]
        self.ingest_core_data(stream_id, core_data)

    def select_columns_to_remove(self):
        columns_to_remove = set()
        for column in self.keys:
            for other_column in self.keys:
                if column != other_column and column in other_column:
                    # the below check ensures that it's a parent column, not just a column with a shared prefix
                    # for instance, we want to remove 'position' if 'position.x' is present, but not if only 'positionx' is present
                    if other_column[len(column)] == '.':
                        columns_to_remove.add(column)
        return list(columns_to_remove)

    def filter_columns(self):
        """Remove columns whose name is a parent of another column's name."""
        columns_to_remove = self.select_columns_to_remove()
        if not columns_to_remove:
            return
        self.dataframe = self.dataframe.drop(columns_to_remove, axis=1)

    def finalize(self):
        return self.dataframe.set_index('time')


def prep_stream_id(stream_id):
    engines = {
        '0': 'GNC',
        '1': 'CDH',
        '2': 'Power',
        '3': 'Thermal'
    }
    split_base = stream_id.split('/')
    prepped_stream_id = f"{split_base[0]}.{engines[split_base[1]]}"
    return prepped_stream_id


class DownloadWorker:
    def __init__(self, download_bar):
        self.download_bar = download_bar
        self.streams: dict[str: StreamManager] = {}
        self.stats = {}
        self.derived_streams: dict[str: StreamManager] = {}
        self.derived_static = {}

    def ingest(self, page, target: dict = None):
        if target is None:
            target = self.streams
        for stream_id, stream_data in page.items():
            if stream_id not in self.streams:
                self.streams[stream_id] = StreamManager(self.download_bar)
            self.streams[stream_id].ingest(stream_id, stream_data)

    def ingest_derived(self, page):
        self.ingest(page, self.derived_streams)

    def finalize(self):
        # process regular and derived series data separately
        for stream_set in [self.streams, self.derived_streams]:
            for stream_manager in stream_set.values():
                # REF 1: https://docs.dask.org/en/stable/dataframe-best-practices.html#repartition-to-reduce-overhead
                stream_manager.dataframe = stream_manager.dataframe.repartition(partition_size="100MB")  # REF 1
                stream_manager.dataframe = stream_manager.dataframe.reset_index(drop=True)
                stream_manager.filter_columns()
                stream_manager.dataframe = stream_manager.dataframe.persist()
                stream_manager.dataframe = stream_manager.dataframe.set_index('time')
        # merge the derived data into the regular data
        for k in self.derived_streams:
            if k not in self.streams:
                self.streams[k] = self.derived_streams[k]
            else:
                self.streams[k].dataframe = self.streams[k].dataframe.join(self.derived_streams[k], how='left')
        # finalize
        for k in self.streams:
            self.streams[k] = self.streams[k].finalize()

    def add_static_data(self, static_data: dict):
        self.derived_static = static_data

    def update_static_data(self, new_static_data: dict):
        for k in new_static_data:
            if k not in self.derived_static:
                self.derived_static[k] = {}
            self.derived_static[k].update(new_static_data[k])

    def finalize_static_data(self, others: "list[DownloadWorker]"):
        for other in others:
            self.update_static_data(other.derived_static)
        return self.derived_static

    def add_stats(self, stats: dict):
        self.stats = stats

    def update_stats(self, new_stats: dict):
        for stream in new_stats:
            if stream not in self.stats:
                self.stats[stream] = {}
            self.stats[stream].update(new_stats[stream])

    def finalize_stats(self, others: "list[DownloadWorker]"):
        for other in others:
            self.update_stats(other.stats)
        return self.stats

    def add_metadata(self, metadata: dict):
        self.metadata = metadata

    def update_metadata(self, new_metadata):
        for k in new_metadata['counts']:
            if k not in self.metadata['counts']:
                self.metadata['counts'][k] = 0
            self.metadata['counts'][k] += new_metadata['counts'][k]

    def finalize_metadata(self, others: "list[DownloadWorker]"):
        for other in others:
            self.update_metadata(other.metadata)
        return self.metadata



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: cosim.proto
# Protobuf Python Version: 5.27.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    2,
    '',
    'cosim.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0b\x63osim.proto\x12\x05\x63osim\"\xec\x01\n\x10SimulationAction\x12\x1e\n\x16\x63luster_handle_address\x18\x01 \x01(\t\x12\x0e\n\x06job_id\x18\x02 \x01(\t\x12\x10\n\x08\x61gent_id\x18\x03 \x01(\t\x12\x1f\n\x17\x65xternal_state_block_id\x18\x04 \x01(\t\x12\x16\n\ttimestamp\x18\x05 \x01(\x01H\x01\x88\x01\x01\x12!\n\x07\x63onsume\x18\x06 \x01(\x0b\x32\x0e.cosim.ConsumeH\x00\x12!\n\x07produce\x18\x07 \x01(\x0b\x32\x0e.cosim.ProduceH\x00\x42\t\n\x07requestB\x0c\n\n_timestamp\"\xe6\x01\n\x12SimulationResponse\x12\x1b\n\x05state\x18\x01 \x01(\x0e\x32\x0c.cosim.State\x12\x11\n\ttimestamp\x18\x02 \x01(\x01\x12\x32\n\x10\x63onsume_response\x18\x03 \x01(\x0b\x32\x16.cosim.ConsumeResponseH\x00\x12\x32\n\x10produce_response\x18\x04 \x01(\x0b\x32\x16.cosim.ProduceResponseH\x00\x12,\n\rfail_response\x18\x05 \x01(\x0b\x32\x13.cosim.FailResponseH\x00\x42\n\n\x08response\"\x18\n\x07\x43onsume\x12\r\n\x05index\x18\x01 \x01(\x05\"/\n\x0f\x43onsumeResponse\x12\r\n\x05index\x18\x01 \x01(\x01\x12\r\n\x05value\x18\x02 \x01(\t\"\'\n\x07Produce\x12\r\n\x05index\x18\x01 \x01(\x05\x12\r\n\x05value\x18\x02 \x01(\t\" \n\x0fProduceResponse\x12\r\n\x05index\x18\x01 \x01(\x05\"\x1f\n\x0c\x46\x61ilResponse\x12\x0f\n\x07message\x18\x01 \x01(\t\"?\n\tAuthorize\x12\x12\n\nauth_token\x18\x01 \x01(\t\x12\x1e\n\x16\x63luster_handle_address\x18\x02 \x01(\t\":\n\x11\x41uthorizeResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x14\n\x0csession_uuid\x18\x02 \x01(\t*!\n\x05State\x12\x0b\n\x07SUCCESS\x10\x00\x12\x0b\n\x07\x46\x41ILURE\x10\x01\x32\x8a\x01\n\x05\x43osim\x12\x44\n\x0eSimulationCall\x12\x17.cosim.SimulationAction\x1a\x19.cosim.SimulationResponse\x12;\n\rAuthorizeCall\x12\x10.cosim.Authorize\x1a\x18.cosim.AuthorizeResponseb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'cosim_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_STATE']._serialized_start=802
  _globals['_STATE']._serialized_end=835
  _globals['_SIMULATIONACTION']._serialized_start=23
  _globals['_SIMULATIONACTION']._serialized_end=259
  _globals['_SIMULATIONRESPONSE']._serialized_start=262
  _globals['_SIMULATIONRESPONSE']._serialized_end=492
  _globals['_CONSUME']._serialized_start=494
  _globals['_CONSUME']._serialized_end=518
  _globals['_CONSUMERESPONSE']._serialized_start=520
  _globals['_CONSUMERESPONSE']._serialized_end=567
  _globals['_PRODUCE']._serialized_start=569
  _globals['_PRODUCE']._serialized_end=608
  _globals['_PRODUCERESPONSE']._serialized_start=610
  _globals['_PRODUCERESPONSE']._serialized_end=642
  _globals['_FAILRESPONSE']._serialized_start=644
  _globals['_FAILRESPONSE']._serialized_end=675
  _globals['_AUTHORIZE']._serialized_start=677
  _globals['_AUTHORIZE']._serialized_end=740
  _globals['_AUTHORIZERESPONSE']._serialized_start=742
  _globals['_AUTHORIZERESPONSE']._serialized_end=800
  _globals['_COSIM']._serialized_start=838
  _globals['_COSIM']._serialized_end=976
# @@protoc_insertion_point(module_scope)



import asyncio
import base64
import itertools
import json
import logging
import os
from typing import Any, Optional, Tuple
from urllib.parse import urlencode

import grpc
from sedaro.sedaro_api_client import SedaroApiClient

from ..utils import serdes
from ..settings import GRPC_MAX_MESSAGE_LENGTH
from . import cosim_pb2, cosim_pb2_grpc

REFRESH_INTERVAL = 60 * 4


class MetadataClientInterceptor(grpc.aio.UnaryUnaryClientInterceptor):
    def __init__(self):
        self.metadata_to_add = None

    def set_uuid(self, uuid: str):
        self.metadata_to_add = [("session-uuid", uuid)]

    def _add_metadata(self, client_call_details):
        metadata = list(client_call_details.metadata) if client_call_details.metadata else []
        if self.metadata_to_add:
            metadata.extend(self.metadata_to_add)
        return client_call_details._replace(metadata=metadata)

    async def intercept_unary_unary(self, continuation, client_call_details, request):
        new_details = self._add_metadata(client_call_details)
        return await continuation(new_details, request)


class CosimClient:
    def __init__(self, grpc_host: str, address: str, job_id: str, host: str, sedaro: SedaroApiClient, insecure: bool = False):
        self.grpc_host = grpc_host
        self.address = address
        self.job_id = job_id
        self.host = host
        self.insecure = insecure

        self.channel: Optional[grpc.aio.Channel] = None
        self.cosim_stub: Optional['cosim_pb2_grpc.CosimStub'] = None
        self.auth_stub: Optional['cosim_pb2_grpc.CosimStub'] = None
        self.sedaro = sedaro
        self._produce_counter = itertools.count(start=1)
        self._consume_counter = itertools.count(start=1)

        self._metadata_interceptor = MetadataClientInterceptor()
        self._refresh_task: Optional[asyncio.Task] = None
        self._stop_refresh = asyncio.Event()

    async def __aenter__(self):
        await self._connect()
        authorized, uuid = await self._authorize()
        self._metadata_interceptor.set_uuid(uuid)
        if not authorized:
            await self._terminate()
            raise Exception("Authentication with CosimClient failed.")

        self._refresh_task = asyncio.create_task(self._refresh_loop())
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self._terminate()

    def _create_channel(self):
        options = [
            ('grpc.max_send_message_length', GRPC_MAX_MESSAGE_LENGTH),
            ('grpc.max_receive_message_length', GRPC_MAX_MESSAGE_LENGTH)
        ]
        
        if self.insecure:
            logging.warning("Using insecure gRPC connection.")
            return grpc.aio.insecure_channel(self.grpc_host, options=options, interceptors=(self._metadata_interceptor,))

        logging.info("Using SSL/TLS for gRPC connection.")
        if os.environ.get("COSIM_TLS_CERTIFICATE"):
            certificate_chain = base64.b64decode(os.environ['COSIM_TLS_CERTIFICATE'])
            credentials = grpc.ssl_channel_credentials(root_certificates=certificate_chain)
            channel = grpc.aio.secure_channel(self.grpc_host, credentials, options=options, interceptors=(self._metadata_interceptor,))
        else:
            credentials = grpc.ssl_channel_credentials()
            channel = grpc.aio.secure_channel(self.grpc_host, credentials, options=options, interceptors=(self._metadata_interceptor,))

        return channel

    async def _connect(self):
        logging.info(f"Connecting to gRPC server at {self.grpc_host}")
        self.channel = self._create_channel()
        await self.channel.channel_ready()
        self.cosim_stub = cosim_pb2_grpc.CosimStub(self.channel)
        self.auth_stub = cosim_pb2_grpc.CosimStub(self.channel)
        logging.info("Connected to gRPC server.")

    async def _refresh_loop(self):
        while not self._stop_refresh.is_set():
            try:
                try:
                    await asyncio.wait_for(self._stop_refresh.wait(), timeout=REFRESH_INTERVAL)
                    break
                except asyncio.TimeoutError:
                    pass

                if not self._stop_refresh.is_set():
                    logging.info("Refreshing authorization token...")
                    authorized, new_uuid = await self._authorize()
                    if authorized:
                        self._metadata_interceptor.set_uuid(new_uuid)
                        logging.info("Authorization token refreshed successfully.")
                    else:
                        logging.error("Failed to refresh authorization token.")
                        self._stop_refresh.set()
                        await self._terminate()
            except Exception as e:
                logging.error(f"Error in refresh loop: {e}")
                continue

    async def _get_auth_token(self) -> Optional[str]:
        """Get a fresh JWT token from the django server."""
        params = {"audience": "SimBed", "permission": "RUN_SIMULATION"}
        url = f"/simulations/jobs/authorization/{self.job_id}?{urlencode(params)}"
        logging.info(f"Getting auth token from {url}")
        return self.sedaro.request.get(url).get('jwt')

    async def _authorize(self) -> Tuple[bool, Optional[str]]:
        jwt_token = await self._get_auth_token()
        if not jwt_token:
            return False, None

        logging.info(f"Authenticating with cosim server.")
        auth_request = cosim_pb2.Authorize(
            auth_token=jwt_token,
            cluster_handle_address=self.address,
        )
        try:
            response: cosim_pb2.AuthorizeResponse = await self.auth_stub.AuthorizeCall(auth_request)
            if response.success:
                logging.info("Authentication successful.")
                return True, response.session_uuid
            else:
                logging.error(f"Authentication failed: {response.message}")
                return False, None
        except grpc.aio.AioRpcError as e:
            logging.error(f"Authentication RPC failed: {e}")
            return False, None

    async def _send_simulation_action(
        self,
        action: cosim_pb2.SimulationAction,
    ) -> Any:
        try:
            logging.debug(f"Sending SimulationAction: {action}")
            response: cosim_pb2.SimulationResponse = await self.cosim_stub.SimulationCall(action)

            if response.state != cosim_pb2.State.SUCCESS:
                logging.error(f"Unexpected response state: {response.state}")
                raise Exception("Unexpected response state.")

            if action.WhichOneof("request") == "consume":
                consume_resp = response.WhichOneof("response")
                if consume_resp != "consume_response":
                    raise Exception("Expected ConsumeResponse.")
                return response.consume_response.value

            elif action.WhichOneof("request") == "produce":
                produce_resp = response.WhichOneof("response")
                if produce_resp != "produce_response":
                    raise Exception("Expected ProduceResponse.")
                return response.produce_response.index

            else:
                raise Exception("Unknown action type.")

        except grpc.aio.AioRpcError as e:
            logging.error(f"Simulation RPC failed: {e}")
            raise ConnectionError(f"Simulation RPC failed: {e}")

    async def _terminate(self):
        """Cleanup resources and stop the refresh loop."""
        if self._refresh_task is not None:
            self._stop_refresh.set()
            await self._refresh_task
            self._refresh_task = None

        if self.channel:
            await self.channel.close()
            logging.info("gRPC channel closed.")
        logging.info("CosimClient terminated.")

    async def produce(
        self,
        external_state_id: str,
        agent_id: str,
        values: Tuple,
        timestamp: Optional[float] = None
    ):
        index = next(self._produce_counter)
        produce_action = cosim_pb2.Produce(
            index=index,
            value=json.dumps({"payload": serdes(values)})
        )
        simulation_action = cosim_pb2.SimulationAction(
            cluster_handle_address=self.address,
            job_id=self.job_id,
            agent_id=agent_id,
            external_state_block_id=external_state_id,
        )
        if timestamp is not None:
            simulation_action.timestamp = timestamp
        simulation_action.produce.CopyFrom(produce_action)

        try:
            response = await self._send_simulation_action(
                simulation_action,
            )
            logging.info(f"Produced message with index {index}: {response}")
            return response
        except Exception as e:
            logging.error(f"Produce operation failed for index {index}: {e}")
            raise e

    async def consume(
        self,
        external_state_id: str,
        agent_id: str,
        timestamp: Optional[float] = None
    ):
        index = next(self._consume_counter)
        consume_action = cosim_pb2.Consume(
            index=index
        )
        simulation_action = cosim_pb2.SimulationAction(
            cluster_handle_address=self.address,
            job_id=self.job_id,
            agent_id=agent_id,
            external_state_block_id=external_state_id,
        )
        if timestamp is not None:
            simulation_action.timestamp = timestamp

        simulation_action.consume.CopyFrom(consume_action)
        try:
            response = await self._send_simulation_action(
                simulation_action,
            )
            logging.info(f"Consumed message with index {index}: {response}")
            return tuple(serdes(v) for v in json.loads(response)["payload"])
        except Exception as e:
            logging.error(f"Consume operation failed for index {index}: {e}")
            raise e



import json
from typing import TYPE_CHECKING, List, Optional, Tuple

import msgpack
import requests

from ...utils import parse_urllib_response

if TYPE_CHECKING:
    from ...sedaro_api_client import SedaroApiClient

class FastFetcherResponse:
    def __init__(self, response: requests.Response):
        self.type = response.headers['Content-Type']

        if self.type == 'application/json':
            self.data = response.text
        elif self.type == 'application/msgpack':
            self.data = response.content
        else:
            raise Exception(
                f"Unexpected MIME type: {self.type}.  Response content: {response.content}. Status Code: {response.status_code}")

        self.status = response.status_code
        self.response = response

    def __getattr__(self, key):
        return self.response[key]

    def parse(self):
        if self.type == 'application/json':
            return parse_urllib_response(self)
        elif self.type == 'application/msgpack':
            return msgpack.unpackb(self.data)
        else:
            raise Exception(
                f"Unexpected MIME type: {self.response.headers['Content-Type']}.  Response content: {self.data}. Status Code: {self.response.status_code}")


class FastFetcher:
    """Accelerated request handler for data page fetching."""

    def __init__(self, sedaro_api: 'SedaroApiClient'):
        self.sedaro_api = sedaro_api

    def get(self, url):
        return FastFetcherResponse(self.sedaro_api.request.requests_lib_get(url))


def _get_metadata(_sedaro: 'SedaroApiClient', sim_id: str = None, num_workers: Optional[int] = None):
    if num_workers is None:
        request_url = f'/data/{sim_id}/metadata'
    else:
        request_url = f'/data/{sim_id}/metadata?numTokens={num_workers}'
    with _sedaro.api_client() as api:
        response = api.call_api(request_url, 'GET', headers={
            'Content-Type': 'application/json',
            'Accept': 'application/json',  # Required for Sedaro firewall
        })
    response_dict = json.loads(response.data)
    return response_dict

def _get_filtered_streams(requested_streams: list, metadata: dict):
    streams_raw = metadata['streams']
    streams_true = {}
    for stream in streams_raw:
        stream_parts = stream.split('.')
        if stream_parts[0] not in streams_true:
            streams_true[stream_parts[0]] = []
        streams_true[stream_parts[0]].append(stream_parts[1])
    filtered_streams = []
    for stream in requested_streams:
        if stream[0] in streams_true:
            if len(stream) == 1:
                for v in streams_true[stream[0]]:
                    filtered_streams.append((stream[0], v))
            else:
                if stream[0] in streams_true:
                    if stream[1] in streams_true[stream[0]]:
                        filtered_streams.append(stream)
    return filtered_streams

def _get_stats_for_sim_id(
    _sedaro: 'SedaroApiClient',
    sim_id: str,
    streams: List[Tuple[str, ...]] = None
):
    request_url = f'/data/{sim_id}/stats/'
    if streams is not None:
        metadata = _get_metadata(_sedaro, sim_id)
        filtered_streams = _get_filtered_streams(streams, metadata)
        encoded_streams = ','.join(['.'.join(x) for x in filtered_streams])
        request_url += f"?streams={encoded_streams}"
    stats = {}

    # get first page
    fast_fetcher = FastFetcher(_sedaro)
    response = fast_fetcher.get(request_url)
    if response.status != 200:
        if response.status == 409: # stats not yet available
            return None, False
        else:
            raise Exception(f"Failed to get stats for sim_id {sim_id}.  Status code: {response.status}.  Response: {response.data}")
    contents = response.parse()
    stats.update(contents['stats'])

    # get additional pages (if applicable)
    while contents['continuationToken'] is not None:
        token = contents['continuationToken']
        request_url = f'/data/{sim_id}/stats/?continuationToken={token}'
        response = fast_fetcher.get(request_url)
        contents = response.parse()
        stats.update(contents['stats'])

    return stats, True



# coding: utf-8

"""
    Sedaro API

     Allows for consumption of Sedaro services. Read more about Sedaro at [docs.sedaro.com](https://docs.sedaro.com).  ### Clients  **Python:** [sedaro](https://pypi.org/project/sedaro/) - This package provides additional functionality on top of the auto-generated OpenAPI client. See the package docs for more information.  ### API Key  To access the Sedaro service via this API, you will need an API key.  You can generate an API key for your account in the Sedaro [Management Console](https://satellite.sedaro.com/account). Once complete, pass the API key in all requests via the `X_API_KEY` HTTP header.  *API keys grant full access to your account and should never be shared. If you think your API key has been compromised, you can revoke it in the [Management Console](https://satellite.sedaro.com/account).*  ### Jupyter Notebooks  For additional examples of how to use this API for modeling and simulation, see our [Mod-Sim Notebooks](https://github.com/sedaro/modsim-notebooks).  ### Community, Support, Discussion  If you have any issues or suggestions, please reach out:  1. Join the Sedaro Community [Slack](https://join.slack.com/t/sedaro-community/shared_invite/zt-1jps4i711-mXy88AZQ9AV7YcEXr8x7Ow) 2. Email us at support@sedarotech.com  ### Known Issues  - Error responses are more specific than what is shown throughout the documentation.  A 4xx or 5xx error will be returned in all error cases.  Only a `200` status indicates success.  See a given error response for additional details.   # noqa: E501

    The version of the OpenAPI document: 4.11.36
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from sedaro_base_client import schemas  # noqa: F401


class ModelRes(
    schemas.DictSchema
):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    A ModelRes Metamodel
    """


    class MetaOapg:
        
        class properties:
            
            
            class blocks(
                schemas.DictSchema
            ):
            
            
                class MetaOapg:
                    
                    
                    class additional_properties(
                        schemas.ComposedSchema,
                    ):
                    
                    
                        class MetaOapg:
                    
                    
                        def __new__(
                            cls,
                            *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                            _configuration: typing.Optional[schemas.Configuration] = None,
                            **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                        ) -> 'additional_properties':
                            return super().__new__(
                                cls,
                                *_args,
                                _configuration=_configuration,
                                **kwargs,
                            )
                
                def __getitem__(self, name: typing.Union[str, ]) -> MetaOapg.additional_properties:
                    # dict_instance[name] accessor
                    return super().__getitem__(name)
                
                def get_item_oapg(self, name: typing.Union[str, ]) -> MetaOapg.additional_properties:
                    return super().get_item_oapg(name)
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[MetaOapg.additional_properties, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                ) -> 'blocks':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            
            
            class index(
                schemas.DictSchema
            ):
            
            
                class MetaOapg:
                    
                    
                    class additional_properties(
                        schemas.ListSchema
                    ):
                    
                    
                        class MetaOapg:
                            items = schemas.StrSchema
                    
                        def __new__(
                            cls,
                            _arg: typing.Union[typing.Tuple[typing.Union[MetaOapg.items, str, ]], typing.List[typing.Union[MetaOapg.items, str, ]]],
                            _configuration: typing.Optional[schemas.Configuration] = None,
                        ) -> 'additional_properties':
                            return super().__new__(
                                cls,
                                _arg,
                                _configuration=_configuration,
                            )
                    
                        def __getitem__(self, i: int) -> MetaOapg.items:
                            return super().__getitem__(i)
                
                def __getitem__(self, name: typing.Union[str, ]) -> MetaOapg.additional_properties:
                    # dict_instance[name] accessor
                    return super().__getitem__(name)
                
                def get_item_oapg(self, name: typing.Union[str, ]) -> MetaOapg.additional_properties:
                    return super().get_item_oapg(name)
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[MetaOapg.additional_properties, list, tuple, ],
                ) -> 'index':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            migrated = schemas.StrSchema
            
            
            class issues(
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    
                    @classmethod
                    @functools.lru_cache()
                    def all_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            Issues,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'issues':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            _relationships = schemas.DictSchema
            
            
            class _blockNames(
                schemas.ListSchema
            ):
            
            
                class MetaOapg:
                    items = schemas.AnyTypeSchema
            
                def __new__(
                    cls,
                    _arg: typing.Union[typing.Tuple[typing.Union[MetaOapg.items, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ]], typing.List[typing.Union[MetaOapg.items, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ]]],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                ) -> '_blockNames':
                    return super().__new__(
                        cls,
                        _arg,
                        _configuration=_configuration,
                    )
            
                def __getitem__(self, i: int) -> MetaOapg.items:
                    return super().__getitem__(i)
            _supers = schemas.DictSchema
            
            
            class type(
                schemas.EnumBase,
                schemas.StrSchema
            ):
                
                @schemas.classproperty
                def MODEL_RES(cls):
                    return cls("ModelRes")
            __annotations__ = {
                "blocks": blocks,
                "index": index,
                "migrated": migrated,
                "issues": issues,
                "_relationships": _relationships,
                "_blockNames": _blockNames,
                "_supers": _supers,
                "type": type,
            }
        additional_properties = schemas.NotAnyTypeSchema
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["blocks"]) -> MetaOapg.properties.blocks: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["index"]) -> MetaOapg.properties.index: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["migrated"]) -> MetaOapg.properties.migrated: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["issues"]) -> MetaOapg.properties.issues: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["_relationships"]) -> MetaOapg.properties._relationships: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["_blockNames"]) -> MetaOapg.properties._blockNames: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["_supers"]) -> MetaOapg.properties._supers: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["type"]) -> MetaOapg.properties.type: ...
    
    def __getitem__(self, name: typing.Union[typing_extensions.Literal["blocks"], typing_extensions.Literal["index"], typing_extensions.Literal["migrated"], typing_extensions.Literal["issues"], typing_extensions.Literal["_relationships"], typing_extensions.Literal["_blockNames"], typing_extensions.Literal["_supers"], typing_extensions.Literal["type"], ]):
        # dict_instance[name] accessor
        return super().__getitem__(name)
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["blocks"]) -> typing.Union[MetaOapg.properties.blocks, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["index"]) -> typing.Union[MetaOapg.properties.index, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["migrated"]) -> typing.Union[MetaOapg.properties.migrated, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["issues"]) -> typing.Union[MetaOapg.properties.issues, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["_relationships"]) -> typing.Union[MetaOapg.properties._relationships, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["_blockNames"]) -> typing.Union[MetaOapg.properties._blockNames, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["_supers"]) -> typing.Union[MetaOapg.properties._supers, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["type"]) -> typing.Union[MetaOapg.properties.type, schemas.Unset]: ...
    
    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["blocks"], typing_extensions.Literal["index"], typing_extensions.Literal["migrated"], typing_extensions.Literal["issues"], typing_extensions.Literal["_relationships"], typing_extensions.Literal["_blockNames"], typing_extensions.Literal["_supers"], typing_extensions.Literal["type"], ]):
        return super().get_item_oapg(name)

    def __new__(
        cls,
        *_args: typing.Union[dict, frozendict.frozendict, ],
        blocks: typing.Union[MetaOapg.properties.blocks, dict, frozendict.frozendict, schemas.Unset] = schemas.unset,
        index: typing.Union[MetaOapg.properties.index, dict, frozendict.frozendict, schemas.Unset] = schemas.unset,
        migrated: typing.Union[MetaOapg.properties.migrated, str, schemas.Unset] = schemas.unset,
        issues: typing.Union[MetaOapg.properties.issues, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
        _relationships: typing.Union[MetaOapg.properties._relationships, dict, frozendict.frozendict, schemas.Unset] = schemas.unset,
        _blockNames: typing.Union[MetaOapg.properties._blockNames, list, tuple, schemas.Unset] = schemas.unset,
        _supers: typing.Union[MetaOapg.properties._supers, dict, frozendict.frozendict, schemas.Unset] = schemas.unset,
        type: typing.Union[MetaOapg.properties.type, str, schemas.Unset] = schemas.unset,
        _configuration: typing.Optional[schemas.Configuration] = None,
    ) -> 'ModelRes':
        return super().__new__(
            cls,
            *_args,
            blocks=blocks,
            index=index,
            migrated=migrated,
            issues=issues,
            _relationships=_relationships,
            _blockNames=_blockNames,
            _supers=_supers,
            type=type,
            _configuration=_configuration,
        )

from sedaro_base_client.model.issues import Issues
